from svv.tree.export.export_solid import (get_branches, get_points, get_radii, get_normals,
                                                polyline_from_points, generate_polylines, generate_tube,
                                                generate_tubes, union_tubes)
from scipy.interpolate import splprep, splev
import numpy as np
import numpy
from copy import deepcopy
import os

def export_spline(tree_connections, extrude_roots=False):
    network_branches = []
    network_points = []
    network_radii = []
    network_normals = []
    network_terminals = []
    if extrude_roots:
        for i in range(len(tree_connections.connected_network)):
            direction = (tree_connections.connected_network[i].data[0, 3:6] - tree_connections.connected_network[i].data[0, 0:3])
            direction = direction / np.linalg.norm(direction)
            root_extension = tree_connections.connected_network[i].data[0, 21] * 4
            start = tree_connections.connected_network[i].data[0, 0:3].copy()
            for j in range(10):
                new_start = start - direction * root_extension * (j + 1)
                if tree_connections.forest.domain(new_start.reshape(1, 3)).flatten() > 0:
                    tree_connections.connected_network[i].data[0, 0:3] = new_start
                    break
    for i in range(len(tree_connections.connected_network)):
        branches = get_branches(tree_connections.connected_network[i].data)
        network_branches.append(branches)
        network_points.append(get_points(tree_connections.connected_network[i].data, branches))
        network_radii.append(get_radii(tree_connections.connected_network[i].data, branches))
        network_normals.append(get_normals(tree_connections.connected_network[i].data, branches))
        terminals = []
        for branch in branches:
            terminals.append(branch[-1])
        network_terminals.append(np.array(terminals))
    # Match the branch terminals to the connection assignments
    reordered_branches = []
    reordered_points = []
    reordered_radii = []
    reordered_normals = []
    for i in range(len(tree_connections.connected_network)):
        tmp_reordered_branches = []
        tmp_reordered_points = []
        tmp_reordered_radii = []
        tmp_reordered_normals = []
        for a in range(len(tree_connections.assignments[i])):
            ind = np.argwhere(network_terminals[i] == tree_connections.assignments[i][a]).flatten()[0]
            tmp_reordered_branches.append(network_branches[i][ind])
            tmp_reordered_points.append(network_points[i][ind])
            tmp_reordered_radii.append(network_radii[i][ind])
            tmp_reordered_normals.append(network_normals[i][ind])
        reordered_branches.append(tmp_reordered_branches)
        reordered_points.append(tmp_reordered_points)
        reordered_radii.append(tmp_reordered_radii)
        reordered_normals.append(tmp_reordered_normals)
    # For trees != 0 the branches, points, radii, and normals need to be reversed
    connection_vessels = deepcopy(tree_connections.vessels)
    for i in range(1, len(reordered_branches)):
        for j in range(len(reordered_branches[i])):
            reordered_branches[i][j] = list(reversed(reordered_branches[i][j]))
            reordered_points[i][j] = list(reversed(reordered_points[i][j]))
            reordered_radii[i][j] = list(reversed(reordered_radii[i][j]))
            reordered_normals[i][j] = list(reversed(reordered_normals[i][j]))
            connection_vessels[i][j] = np.flip(connection_vessels[i][j], axis=0)
            proximal_points = connection_vessels[i][j][:, 0:3].copy()
            distal_points = connection_vessels[i][j][:, 3:6].copy()
            connection_vessels[i][j][:, 0:3] = distal_points
            connection_vessels[i][j][:, 3:6] = proximal_points
    # Take intermediate points along the connecting vessels
    connection_points = []
    connection_radii = []
    connection_normals = []
    for i in range(len(connection_vessels)):
        tmp_tree_points = []
        tmp_tree_radii = []
        tmp_tree_normals = []
        for j in range(len(connection_vessels[i])):
            tmp_branch_points = []
            tmp_branch_radii = []
            tmp_branch_normals = []
            for k in range(connection_vessels[i][j].shape[0]):
                if i > 1:
                    tmp_branch_points.append(connection_vessels[i][j][k, 0:3].tolist())
                one_fourth = connection_vessels[i][j][k, 0:3] * (3 / 4) + connection_vessels[i][j][k, 3:6] * (1 / 4)
                tmp_branch_points.append(one_fourth.tolist())
                mid = connection_vessels[i][j][k, 0:3] * (1 / 2) + connection_vessels[i][j][k, 3:6] * (1 / 2)
                tmp_branch_points.append(mid.tolist())
                three_fourths = connection_vessels[i][j][k, 0:3] * (1 / 4) + connection_vessels[i][j][k, 3:6] * (3 / 4)
                tmp_branch_points.append(three_fourths.tolist())
                if i > 1:
                    tmp_branch_radii.append(connection_vessels[i][j][k, 6])
                tmp_branch_radii.append(connection_vessels[i][j][k, 6])
                tmp_branch_radii.append(connection_vessels[i][j][k, 6])
                tmp_branch_radii.append(connection_vessels[i][j][k, 6])
                normal = connection_vessels[i][j][k, 3:6] - connection_vessels[i][j][k, 0:3]
                normal = normal / np.linalg.norm(normal)
                if i > 1:
                    tmp_branch_normals.append(normal.tolist())
                tmp_branch_normals.append(normal.tolist())
                tmp_branch_normals.append(normal.tolist())
                tmp_branch_normals.append(normal.tolist())
            tmp_tree_points.append(tmp_branch_points)
            tmp_tree_radii.append(tmp_branch_radii)
            tmp_tree_normals.append(tmp_branch_normals)
        connection_points.append(tmp_tree_points)
        connection_radii.append(tmp_tree_radii)
        connection_normals.append(tmp_tree_normals)
    # Combine the reordered points, radii, and normals with the connection points, radii, and normals
    all_points = []
    all_radii = []
    all_normals = []
    # Because the first and second vessels are lofted together
    # to ensure smoothness, the first vessel is not included
    for i in range(1, len(tree_connections.connected_network)):
        network_vessel_points = []
        network_vessel_radii = []
        network_vessel_normals = []
        for j in range(len(connection_points[i])):
            vessel = []
            radii = []
            normals = []
            if i == 1:
                vessel.extend(reordered_points[0][j])
                vessel.extend(connection_points[0][j])
                vessel.extend(connection_points[i][j])
                vessel.extend(reordered_points[i][j])
                radii.extend(reordered_radii[0][j])
                radii.extend(connection_radii[0][j])
                radii.extend(connection_radii[i][j])
                radii.extend(reordered_radii[i][j])
                normals.extend(reordered_normals[0][j])
                normals.extend(connection_normals[0][j])
                normals.extend(connection_normals[i][j])
                normals.extend(reordered_normals[i][j])
            else:
                vessel.extend(connection_points[i][j])
                vessel.extend(reordered_points[i][j])
                radii.extend(connection_radii[i][j])
                radii.extend(reordered_radii[i][j])
                normals.extend(connection_normals[i][j])
                normals.extend(reordered_normals[i][j])
            network_vessel_points.append(vessel)
            network_vessel_radii.append(radii)
            network_vessel_normals.append(normals)
        all_points.append(network_vessel_points)
        all_radii.append(network_vessel_radii)
        all_normals.append(network_vessel_normals)
    # Now we have complete sets of points, radii, and normals
    # defining the solids for each vessel
    # Create the solid
    interp_xyz = []
    interp_radii = []
    interp_normals = []
    for i in range(len(all_points)):
        network_xyz = []
        network_r = []
        network_n = []
        network_xyzr = []
        for j in range(len(all_points[i])):
            p = np.array(all_points[i][j]).T
            r = np.array(all_radii[i][j]).T
            n = np.array(all_normals[i][j]).T
            if p.shape[1] == 2:
                network_xyz.append(splprep(p, k=1, s=0))
                rr = numpy.vstack((network_xyz[-1][1], r))
                network_r.append(splprep(rr, k=1, s=0))
                xyzr = numpy.vstack((p, r))
                network_xyzr.append(splprep(xyzr, k=1, s=0))
                # interp_n.append(splprep(n, k=1, s=0))
            elif p.shape[1] == 3:
                interp_xyz.append(splprep(p, k=2, s=0))
                rr = numpy.vstack((network_xyz[-1][1], r))
                network_r.append(splprep(rr, k=1, s=0))
                xyzr = numpy.vstack((p, r))
                network_xyzr.append(splprep(xyzr, k=2, s=0))
                # interp_n.append(splprep(n, k=2, s=0))
            else:
                network_xyz.append(splprep(p, s=0))
                rr = numpy.vstack((network_xyz[-1][1], r))
                network_r.append(splprep(rr, k=1, s=0))
                xyzr = numpy.vstack((p, r))
                network_xyzr.append(splprep(xyzr, s=0))
                # interp_n.append(splprep(n, s=0))
        interp_xyz.append(network_xyz)
        interp_radii.append(network_r)
        interp_normals.append(network_n)
    return interp_xyz, interp_radii, interp_normals, all_points, all_radii, all_normals


def write_splines(ALL_POINTS, ALL_RADII, spline_sample_points=100, seperate=False, write_splines=True):
    ALL_SPLINES = []
    for network in range(len(ALL_POINTS)):
        network_splines = []
        if write_splines:
            spline_file = open(os.getcwd() + os.sep + "network_{}_b_splines.txt".format(network), "w+")
        for vessel in range(len(ALL_POINTS[network])):
            pt_array = np.array(ALL_POINTS[network][vessel])
            r_array = np.array(ALL_RADII[network][vessel]).reshape(-1, 1)
            pt_r_combined = deepcopy(np.hstack((pt_array, r_array)).T)
            vessel_ctr = splprep(pt_r_combined, s=0)
            def vessel_spline(t, ctr=vessel_ctr):
                return splev(t, ctr[0])
            #vessel_spline = deepcopy(lambda t: splev(t, deepcopy(vessel_ctr[0])))
            network_splines.append(deepcopy(vessel_spline))
            if write_splines:
                spline_file.write('Vessel: {}, Number of Points: {}\n\n'.format(vessel, spline_sample_points))
                t = np.linspace(0, 1, num=spline_sample_points)
                data = deepcopy(vessel_spline(t))
                for k in range(spline_sample_points):
                    if k > spline_sample_points // 2:
                        label = 1
                    else:
                        label = 0
                    if seperate:
                        spline_file.write(
                                    '{}, {}, {}, {}, {}\n'.format(data[0][k], data[1][k], data[2][k], data[3][k],
                                                                  label))
                    else:
                        spline_file.write(
                                    '{}, {}, {}, {}\n'.format(data[0][k], data[1][k], data[2][k], data[3][k]))
                spline_file.write('\n')
        spline_file.close()
        ALL_SPLINES.append(network_splines)
        return ALL_SPLINES


# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
The module contains utilities used to read files, calculate centers, etc.
"""

import os.path
import vtk

import logging
from .manage import get_logger_name
logger = logging.getLogger(get_logger_name())

SurfaceFileFormats = ["vtk", "vtp"]


def read_surface(file_name, file_format="vtp", datatype=None):
    """
    Read surface geometry from a file.

    Args:
        file_name (str): Path to input file.
        file_format (str): File format (.vtp, .stl, etc.).
        datatype (str): Additional parameter for vtkIdList objects.

    Returns:
        polyData (vtkSTL/vtkPolyData/vtkXMLStructured/
                    vtkXMLRectilinear/vtkXMLPolydata/vtkXMLUnstructured/
                    vtkXMLImage/Tecplot): Output data.
    """

    # Check if file exists
    if not os.path.exists(file_name):
        raise RuntimeError("Could not find file: %s" % file_name)

    # Get reader
    if file_format == 'stl':
        reader = vtk.vtkSTLReader()
        reader.MergingOn()
    elif file_format == 'vtk':
        reader = vtk.vtkPolyDataReader()
    elif file_format == 'vtp':
        reader = vtk.vtkXMLPolyDataReader()
    elif file_format == 'vts':
        reader = vtk.vtkXMinkorporereLStructuredGridReader()
    elif file_format == 'vtr':
        reader = vtk.vtkXMLRectilinearGridReader()
    elif file_format == 'vtu':
        reader = vtk.vtkXMLUnstructuredGridReader()
    elif file_format == "vti":
        reader = vtk.vtkXMLImageDataReader()
    elif file_format == "np" and datatype == "vtkIdList":
        result = np.load(filename).astype(np.int)
        id_list = vtk.vtkIdList()
        id_list.SetNumberOfIds(result.shape[0])
        for i in range(result.shape[0]):
            id_list.SetId(i, result[i])
        return id_list
    else:
        raise RuntimeError('Unknown file type %s' % file_format)

    # Read surface geometry.
    reader.SetFileName(file_name)
    reader.Update()
    polydata = reader.GetOutput()
    polygons = polydata.GetPolys()
    num_polys = polygons.GetNumberOfCells()
    return polydata


def write_polydata(file_name, data, datatype=None):

    """ Write the given VTK object to a file.

    Args:
        file_name (str): The name of the file to write.
        data (vtkDataObject): Data to write.
        datatype (str): Additional parameter for vtkIdList objects.
    """
    # Check filename format.
    file_ext = file_name.split(".")[-1]

    if file_ext == '':
        raise RuntimeError('The file does not have an extension')

    # Get writer.
    if file_ext == 'vtp':
        writer = vtk.vtkXMLPolyDataWriter()
    elif file_ext == 'vtu':
        writer = vtk.vtkXMLUnstructuredGridWriter()
    else:
        raise RuntimeError('Unknown file type %s' % file_ext)

    # Set file name and the data to write.
    writer.SetFileName(file_name)
    writer.SetInputData(data)
    writer.Update()

    # Write the data.
    writer.Write()


def read_polydata(file_name, datatype=None):
    """ Read a vtp file.

    Args:
        file_name (str): The input polydata file.
        datatype (str): Additional parameter for vtkIdList objects.

    Returns:
        polydata (vtkDataObject): Geometry read in from the input file.
    """

    # Check if file exists.
    if not os.path.exists(file_name):
        raise RuntimeError("Could not find file: %s" % file_name)

    # Check filename format.
    file_ext = file_name.split(".")[-1]

    if file_ext == '':
        raise RuntimeError('The file does not have an extension')

    if file_ext == 'vtp':
        reader = vtk.vtkXMLPolyDataReader()
    else:
        raise RuntimeError('Unknown file type %s' % file_ext)

    # Read the vtkPolyData object.
    reader.SetFileName(file_name)
    reader.Update()
    polydata = reader.GetOutput()

    return polydata




# test_base_curve.py

import pytest
import numpy as np

from svv.forest.connect.base import BaseCurve

def test_cannot_instantiate_base_curve():
    """
    Attempting to instantiate an ABC (BaseCurve) directly should fail
    because it has abstract methods and properties.
    """
    with pytest.raises(TypeError):
        _ = BaseCurve()

# ----------------------------------------------------------------------------
# Mock subclass that provides minimal implementations of abstract methods.
# ----------------------------------------------------------------------------
class MockCurve(BaseCurve):
    """
    A minimal subclass of BaseCurve that just implements the abstract methods
    enough to allow instantiation and testing of base-class functionality.
    """

    def evaluate(self, t_values):
        # Return a straight line, e.g. y = x, in 2D
        t_values = np.atleast_1d(t_values)
        return np.column_stack((t_values, t_values))

    def derivative(self, t_values, order=1):
        # Derivative of y=x line is (1,1), second derivative is (0,0), etc.
        t_values = np.atleast_1d(t_values)
        if order == 1:
            # shape = (len(t_values), dimension)
            return np.tile([1.0, 1.0], (len(t_values), 1))
        else:
            # higher derivatives are zero
            return np.zeros((len(t_values), 2))

    def roc(self, t_values):
        # Radius of curvature for a line is infinite. Return large number.
        t_values = np.atleast_1d(t_values)
        return np.full(t_values.shape, 1e9)

    def torsion(self, t_values):
        # In 2D line, torsion = 0
        t_values = np.atleast_1d(t_values)
        return np.zeros_like(t_values, dtype=float)

    def arc_length(self, t_start=0.0, t_end=1.0, num_points=100):
        # For line y=x from t_start to t_end, the length is sqrt(2)*(t_end - t_start)
        return np.sqrt(2) * (t_end - t_start)

    @property
    def dimension(self):
        return 2  # This mock curve is in 2D

# ----------------------------------------------------------------------------
# Test the mock subclass (which indirectly tests some base-class behaviors).
# ----------------------------------------------------------------------------

def test_mock_curve_instantiation():
    """
    Instantiating MockCurve should succeed because
    it implements all abstract methods.
    """
    curve = MockCurve()
    assert isinstance(curve, BaseCurve)
    assert curve.dimension == 2

def test_mock_curve_evaluate():
    """
    Check that the mock curve returns the expected (x, x) line.
    """
    curve = MockCurve()
    t_values = [0, 0.5, 1.0]
    pts = curve.evaluate(t_values)
    expected = np.array([[0.0, 0.0],
                         [0.5, 0.5],
                         [1.0, 1.0]])
    np.testing.assert_allclose(pts, expected, atol=1e-8)

def test_mock_curve_derivative():
    """
    For a line y=x, the first derivative is (1,1),
    second derivative is (0,0), etc.
    """
    curve = MockCurve()
    t_values = [0, 0.25, 0.75, 1]
    d1 = curve.derivative(t_values, order=1)
    np.testing.assert_allclose(d1, [[1,1]]*len(t_values), atol=1e-8)
    d2 = curve.derivative(t_values, order=2)
    np.testing.assert_allclose(d2, [[0,0]]*len(t_values), atol=1e-8)

def test_mock_curve_roc():
    """
    For the line mock, roc() returns a large constant (representing infinity).
    """
    curve = MockCurve()
    t_vals = np.linspace(0, 1, 5)
    roc_vals = curve.roc(t_vals)
    assert np.all(roc_vals > 1e6), "ROC for a line should be large (infinite)."

def test_mock_curve_torsion():
    """
    Torsion in 2D line should be 0.
    """
    curve = MockCurve()
    t_vals = np.linspace(0, 1, 5)
    torsion_vals = curve.torsion(t_vals)
    np.testing.assert_allclose(torsion_vals, 0.0)

def test_mock_curve_arc_length():
    """
    Arc length for line y=x from 0 to 1 is sqrt(2).
    """
    curve = MockCurve()
    length = curve.arc_length(t_start=0, t_end=1, num_points=50)
    np.testing.assert_allclose(length, np.sqrt(2), atol=1e-6)

def test_mock_curve_bounding_box():
    """
    bounding_box is a base-class method that uses evaluate internally.
    The line from t=0 to t=1 goes from (0,0) to (1,1).
    """
    curve = MockCurve()
    min_pt, max_pt = curve.bounding_box(num_samples=5)
    np.testing.assert_allclose(min_pt, [0, 0])
    np.testing.assert_allclose(max_pt, [1, 1])

def test_mock_curve_is_closed():
    """
    is_closed checks whether the start point and end point are within tolerance.
    The line from (0,0) to (1,1) is not closed.
    """
    curve = MockCurve()
    assert curve.is_closed() == False

def test_transform_not_implemented():
    """
    The base transform method raises NotImplementedError
    unless a subclass overrides it.
    """
    curve = MockCurve()
    with pytest.raises(NotImplementedError):
        curve.transform(np.eye(3))



# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module handles file i/o for the SimVascular 1D solver (https://github.com/SimVascular/oneDSolver).
"""
import pdb
from os import path
import logging
import re
from .manage import get_logger_name
from .parameters import OutflowBoundaryConditionType, MaterialModel

import numpy as np
from collections import OrderedDict

import vtk.util.numpy_support as nps
from vtk import vtkIdList
from vtk import vtkPoints, vtkLine, vtkCellArray, vtkPolyData, vtkXMLPolyDataWriter
from .utils import SurfaceFileFormats, read_polydata, write_polydata
from .io_headers import Headers
from collections import namedtuple

# Define flow data tuple.
FlowData = namedtuple('FlowData', 'time flow')


def read_inflow_file(mesh, params):
    """ Read the inflow file.

    The file can be formatted as space- or comma-separated value pairs.

    Example:
        <time1>  <flow1>
        <time2>  <flow2>
        ...
        <timeN>  <flowN>
    """
    mesh.logger.info("Read inflow BC ...")
    inflow_data = []
    inflow_file = params.inflow_input_file

    try:
        with open(inflow_file, "r") as ofile:
            for line in ofile:
                if line.strip() == '':
                    continue
                values = re.split("[, ]+", line.strip())
                inflow_data.append(FlowData(time=float(values[0]), flow=float(values[1])))

    except Exception as e:
        msg = "The inflow file is in the wrong format, expecting space- or comma-separated value pairs.\n"
        mesh.logger.error(msg)
        raise RuntimeError(str(e))

    mesh.inflow_data = inflow_data


def read_outlet_face_names(mesh, params):
    """ Read in outlet face names file.
    """
    outlet_face_names = []
    mesh.logger.info("Read model outlet face names from: %s" % params.outlet_face_names_file)
    with open(params.outlet_face_names_file) as file:
        for line in file:
            outlet_face_names.extend(line.splitlines())
    mesh.logger.info("Number of model outlet faces names: %d" % len(outlet_face_names))
    return outlet_face_names


def write_mesh(mesh, params):
    """
    Write the 1D mesh to a VTK VTP format file.
    Args:
        mesh: Mesh object
        params: 1d model parameters
    """
    output_dir = params.output_directory
    file_name = path.join(output_dir, params.mesh_output_file)

    # add points and lines
    points = vtkPoints()
    lines = vtkCellArray()
    i = 0
    for branch in mesh.point_data['coord'].values():
        for j, point in enumerate(branch):
            points.InsertNextPoint(params.lcoef * point)
            if j > 0:
                line = vtkLine()
                line.GetPointIds().SetId(0, i + j - 1)
                line.GetPointIds().SetId(1, i + j)
                lines.InsertNextCell(line)
        i += len(branch)

    # create polydata
    polydata = vtkPolyData()
    polydata.SetPoints(points)
    polydata.SetLines(lines)
    polydata.Modified()

    # add point and cell arrays to output
    output = {'point': mesh.point_data, 'cell': mesh.cell_data}
    for field, out in output.items():
        for name, data in out.items():
            array = []
            for branch in data.values():
                for v in branch:
                    array += [v]
            add_array(polydata, field, array, name)

    # write the VTP file
    writer = vtkXMLPolyDataWriter()
    writer.SetFileName(file_name)
    writer.SetInputData(polydata)
    writer.Write()


def write_solver_file(mesh, params):
    """ Write a solver input file.
    """
    # print("Write solver file.")
    mesh.logger.info("Write solver file.")
    output_dir = params.output_directory
    file_name = path.join(output_dir, params.solver_output_file)
    model_name = params.model_name
    sp = mesh.space
    # print("Solver file %s" % file_name)

    # Open file
    ofile = mesh.Open(file_name, "w")

    # write header
    header = Headers().model(model_name)
    write_solver_section_header(mesh, ofile, header)

    # Write node section.
    write_solver_nodes(mesh, ofile, params)

    # Write joint section.
    write_solver_joints(mesh, ofile, params)

    # Write segment section.
    write_solver_segments(mesh, ofile, params)

    # Write SolverOptions section.
    write_solver_options(mesh, ofile, params)

    # Write material section.
    write_solver_material(mesh, ofile, params)

    # Write output section.
    write_solver_output(mesh, ofile, params)

    ofile.close()


def write_solver_nodes(mesh, ofile, params):
    """ Write a solver input file nodes section.
    """
    header = Headers().node

    ofile.writeln(mesh.solver_file_msg)
    write_solver_section_header(mesh, ofile, header)

    lcoef = params.lcoef
    sp = mesh.space

    # loop branches
    for ids, nodes in zip(mesh.point_data['id'].values(), mesh.point_data['coord'].values()):
        # loop nodes
        for i, node in zip(ids, nodes):
            ofile.writeln("NODE " + str(i) + sp + sp.join(str(lcoef * node[j]) for j in range(3)))


def write_solver_joints(mesh, ofile, params):
    """ Write a solver input file joints section.
    """
    header1 = Headers().joint1
    header2 = Headers().joint2

    # Joint header.
    ofile.writeln(mesh.solver_file_msg)
    write_solver_section_header(mesh, ofile, header1)

    # JointInlet and JointOutlet header.
    ofile.writeln(mesh.solver_file_msg)
    write_solver_section_header(mesh, ofile, header2)

    sp = mesh.space
    seg_connectivity = mesh.seg_connectivity
    seg_rear = mesh.seg_rear

    for i in range(0, len(seg_connectivity)):
        joint = "JOINT J" + str(i) + sp + str(seg_rear[i])
        jin = "IN" + str(i)
        jout = "OUT" + str(i)
        ofile.writeln(joint + sp + jin + sp + jout)
        ofile.writeln("JOINTINLET IN" + str(i) + sp + "1 " + str(seg_connectivity[i][0]))
        ofile.write("JOINTOUTLET OUT" + str(i) + sp + str(len(seg_connectivity[i]) - 1))
        for j in range(1, len(seg_connectivity[i])):
            ofile.write(sp + str(seg_connectivity[i][j]))
        ofile.write("\n\n")


def write_solver_segments(mesh, ofile, params):
    """ Write a solver input file joints section.
    """
    header = Headers().segment
    write_solver_section_header(mesh, ofile, header)
    mesh.logger.info("Write solver segment section ...")

    sp = mesh.space

    uniform_bc = params.uniform_bc
    uniform_material = params.uniform_material
    dx = params.element_size
    min_num_elems = params.min_num_elems
    inflow_data = mesh.inflow_data

    mesh.logger.info("Uniform BC: %s" % uniform_bc)
    for t in set(mesh.bc_type.values()):
        mesh.logger.info("Outflow BC: %s" % t)
    mesh.num_elements = 0

    assert uniform_material, 'non-uniform material not implemented'

    # loop branches
    for (branch, ids), lengths, stenosis, seg_1d in zip(mesh.cell_data['id'].items(), mesh.cell_data['length'].values(), mesh.cell_data['stenosis'].values(), mesh.cell_data['1d_seg'].values()):
        # loop nodes
        for i, (j, length, sten, numfe) in enumerate(zip(ids, lengths, stenosis, seg_1d)):
            if uniform_material:
                matname = "MAT1"
            else:
                matname = "MAT_group" + str(j)

            mesh.num_elements += numfe

            p0 = mesh.point_data['id'][branch][i]
            p1 = mesh.point_data['id'][branch][i + 1]
            a0, a1 = mesh.cell_data['area'][branch][i]

            seg_out = "SEGMENT" + sp + "branch" + str(branch) + "_seg" + str(i) + sp + str(j) + sp +\
                      str(length) + sp + str(numfe) + sp + str(p0) + sp +\
                      str(p1) + sp + str(a0) + sp + str(a1) + sp +\
                      "0.0 " + matname

            if (len(stenosis) == 3 and i == 1) or (len(stenosis) > 3 and i > 0):
                seg_out += " STENOSIS 0.0 " + str(j - 1) + " 0 "
            else:
                seg_out += " NONE 0.0 0 0 "

            ofile.write(seg_out)

            if j in mesh.terminal:
                if uniform_bc:
                    outflow_bc = OutflowBoundaryConditionType.RCR.upper()
                    ofile.writeln(outflow_bc + " " + outflow_bc + "_1")
                    msg = "While writing solver segments encountered: group_terminal[seg_list[i]] == 1"
                    mesh.logger.error(msg)
                    raise RuntimeError(msg)
                else:
                    bc_id = mesh.terminal.index(j)
                    bc_name = list(mesh.outlet_face_names_index.keys())[bc_id]
                    bc_type = mesh.bc_type[bc_name].upper()
                    ofile.writeln(bc_type + " " + bc_type + "_" + str(bc_id))
            else:
                ofile.writeln("NOBOUND NONE")

    ofile.writeln("")
    ofile.writeln("")

    if uniform_bc:
        ofile.writeln("DATATABLE " + outflow_bc_uc + "_1 LIST")
        ofile.writeln(sp)
        ofile.writeln("ENDDATATABLE")
        ofile.writeln("")
    else:
        for i, name in enumerate(mesh.outlet_face_names_index.keys()):
            bc_val = mesh.bc_map[name]
            bc_type = mesh.bc_type[name]
            ofile.writeln("DATATABLE " + bc_type.upper() + "_" + str(i) + " LIST")

            if bc_type == OutflowBoundaryConditionType.RCR or bc_type == OutflowBoundaryConditionType.RESISTANCE:
                for j in bc_val:
                    ofile.writeln("0.0 " + str(j))

            elif bc_type == OutflowBoundaryConditionType.CORONARY:
                # write parameters
                for v in ['Ra1', 'Ra2', 'Ca', 'Cc', 'Rv1', 'P_v']:
                    ofile.writeln("-1.0 " + str(bc_val['var'][v]))

                # write time / pressure pairs
                for t, p in zip(bc_val['time'], bc_val['pressure']):
                    ofile.writeln(str(t) + sp + str(p))

            ofile.writeln("ENDDATATABLE")
            ofile.writeln("")

    ofile.writeln("")
    ofile.writeln("")
    ofile.writeln("DATATABLE INFLOW LIST")

    if not inflow_data:
        ofile.writeln("Copy and paste inflow data here.")
    else:
        for value in inflow_data:
            ofile.writeln(" %f %f" % (value.time, value.flow))

    ofile.writeln("ENDDATATABLE")
    ofile.writeln("")
    ofile.writeln("")


def write_solver_options(mesh, ofile, params):
    """ Write a solver input file options section.
    """
    header = Headers().solveroptions
    write_solver_section_header(mesh, ofile, header)

    sp = mesh.space
    time_step = params.time_step
    num_time_steps = params.num_time_steps
    save_data_freq = params.save_data_freq

    ofile.writeln("SOLVEROPTIONS " + str(time_step) + " " + str(save_data_freq) + " " + str(
        num_time_steps) + " 2 INFLOW FLOW 1.0e-5 1 1")
    ofile.writeln("")


def write_solver_material(mesh, ofile, params):
    """ Write a solver input file material section.
    """
    header = Headers().material
    write_solver_section_header(mesh, ofile, header)

    sp = mesh.space
    mattype = params.material_model
    density = params.density
    viscosity = params.viscosity

    if mattype == MaterialModel.OLUFSEN:
        k1 = params.olufsen_material_k1
        k2 = params.olufsen_material_k2
        k3 = params.olufsen_material_k3
        pressure = params.olufsen_material_pressure
        exponent = params.olufsen_material_exponent
    elif mattype == MaterialModel.LINEAR:
        k1 = params.linear_material_ehr
        k2 = 0.0
        k3 = 0.0
        pressure = params.linear_material_pressure
        exponent = 1.0

    uniform_material = params.uniform_material

    if uniform_material:
        ofile.writeln("MATERIAL MAT1 " + mattype + sp + str(density) + sp + str(viscosity) + sp +
                      str(pressure) + sp + str(exponent) + sp + str(k1) + sp + str(k2) + sp + str(k3))
    else:
        msg = 'Non-uniform material not implemented'
        mesh.logger.error(msg)
        raise RuntimeError(msg)


def write_solver_output(mesh, ofile, params):
    """ Write a solver input file output section.
    """
    header = Headers().output
    ofile.writeln("")
    write_solver_section_header(mesh, ofile, header)

    outputformat = params.outputformat
    ofile.writeln("OUTPUT " + outputformat)


def write_solver_section_header(mesh, ofile, header):
    """ Write a solver input file section header.
    """
    endl = mesh.endl
    hdr = endl.join(line for line in header)
    ofile.writeln(hdr)


def add_array(data, identifier, array, name):
    """
    Add array to vtkPolyData
    Args:
        data: vtkPolyData object
        identifier: array association (point or cell)
        array: scalar list or numpy array
        name: display name of array
    """
    out_array = nps.numpy_to_vtk(np.array(array))
    out_array.SetName(name)
    if identifier == 'point':
        data.GetPointData().AddArray(out_array)
    elif identifier == 'cell':
        data.GetCellData().AddArray(out_array)
    else:
        raise ValueError('Unknown identifier ' + identifier + '. Choose point or cell.')


def read_variable_outflow_bcs(params):
    """
    Read in data for variable flow boundary conditions from file (RCR, resistance, coronary)
    """
    bc_map = {}
    bc_type = {}
    for outflow_bc_sv in params.outflow_bc_type:
        # filename for boundary condition type
        bc_file = path.join(params.outflow_bc_file, outflow_bc_sv)
        if not path.exists(bc_file):
            raise RuntimeError('Boundary condition file not found: ' + bc_file)

        # onedsolver name for boundary condition type
        bc_in = path.basename(outflow_bc_sv)
        if bc_in not in OutflowBoundaryConditionType.SV_TO_ONED:
            raise RuntimeError('Unknown bounadry condition type: ' + bc_in)
        outflow_bc = OutflowBoundaryConditionType.SV_TO_ONED[bc_in]

        # switch between boundary condition types resistance, RCR, coronary
        if outflow_bc == OutflowBoundaryConditionType.RESISTANCE:
            try:
                with open(bc_file) as rfile:
                    for line in rfile:
                        split = line.strip().split(' ')
                        face_name = split[0]

                        # resistance
                        value = []
                        value += [split[1]]

                        # check if reference pressure is given
                        if len(split) == 2:
                            value += [0.0]
                        elif len(split) == 3:
                            value += [split[2]]
                        else:
                            msg = "The resistance file is in the wrong format, expecting face name / value pairs."
                            raise RuntimeError(msg)

                        bc_map[face_name] = value
                        bc_type[face_name] = outflow_bc

            except Exception as e:
                msg = "The resistance file is in the wrong format, expecting face name / value pairs.\n"
                raise RuntimeError(msg)

        elif outflow_bc == OutflowBoundaryConditionType.RCR:
            with open(bc_file) as rfile:
                keyword = rfile.readline()
                while True:
                    tmp = rfile.readline()
                    if tmp == keyword:
                        RCRval = []
                        face_name = rfile.readline().strip()
                        RCRval.append(float(rfile.readline()))
                        RCRval.append(float(rfile.readline()))
                        RCRval.append(float(rfile.readline()))

                        # read reference pressure
                        p0 = float(rfile.readline().strip().split(' ')[1])
                        p1 = float(rfile.readline().strip().split(' ')[1])
                        assert p0 == p1, 'Cannot handle time-dependent reference pressure'
                        RCRval.append(float(p1))

                        bc_map[face_name] = RCRval
                        bc_type[face_name] = outflow_bc
                    if len(tmp) == 0:
                        break

        elif outflow_bc == OutflowBoundaryConditionType.CORONARY:
            with open(bc_file) as rfile:
                keyword = rfile.readline()
                tmp = rfile.readline()
                while True:
                    if tmp == keyword:
                        face_name = rfile.readline().strip()

                        # read coronary parameters
                        coronary_var_sv = OrderedDict()
                        var_names = ['q0', 'q1', 'q2', 'p0', 'p1', 'p2', 'b0', 'b1', 'b2', 'dQinidT', 'dPinidT']
                        for name in var_names:
                            try:
                                coronary_var_sv[name] = float(rfile.readline())
                            except Exception:
                                raise RuntimeError('Wrong format of coronary boundary conditions')

                        # convert SimVascular convention to oneD convention
                        coronary_var = coronary_sv_to_oned(coronary_var_sv)

                        # read ventricular time and pressure pairs
                        time = []
                        pressure = []
                        while True:
                            tmp = rfile.readline()
                            # check if end of table
                            if tmp == keyword or len(tmp) == 0:
                                break

                            # split line in time, pressure
                            try:
                                t, p = tmp.split()
                            except Exception:
                                raise RuntimeError('Wrong format of coronary boundary conditions')

                            time += [float(t)]
                            pressure += [float(p)]

                        coronary = {'var': coronary_var, 'time': time, 'pressure': pressure}

                        bc_map[face_name] = coronary
                        bc_type[face_name] = outflow_bc
                    if len(tmp) == 0:
                        break

    return bc_map, bc_type


def coronary_sv_to_oned(bc):
    """
    Convert format of coronary boundary condition parameters from svSimVascular to svOneDSolver
    """
    # unpack constants
    p1, p2, q0, q1, q2, b1 = (bc['p1'], bc['p2'], bc['q0'], bc['q1'], bc['q2'], bc['b1'])
    Rv_micro = 0

    # build system of equations (obtained from analytically solving for constants as defined in paper)
    # see H. J. Kim et al. "Patient-Specific Modeling of Blood Flow and Pressure in Human Coronary Arteries", p. 3198
    Ra = q2 / p2
    Ra_micro = (p1 ** 2 * q2 ** 2 - 2 * p1 * p2 * q1 * q2 + p2 ** 2 * q1 ** 2) / (p2 * (- q2 * p1 ** 2 + q1 * p1 * p2 - q0 * p2 ** 2 + q2 * p2))
    Ca = -p2**2/(p1*q2 - p2*q1)
    Cim = (- q2*p1**2 + q1*p1*p2 - q0*p2**2 + q2*p2)**2 / ((p1*q2 - p2*q1)*(p1**2*q0*q2 - p1*p2*q0*q1 - p1*q1*q2 + p2**2*q0**2 - 2*p2*q0*q2 + p2*q1**2 + q2**2))
    Rv = -(p1**2*q0*q2 - p1*p2*q0*q1 - p1*q1*q2 + p2**2*q0**2 - 2*p2*q0*q2 + p2*q1**2 + q2**2) / (- q2*p1**2 + q1*p1*p2 - q0*p2**2 + q2*p2)

    # check equation residuals
    res = [p1 - (Ra_micro * Ca + (Rv + Rv_micro) * (Ca + Cim)),
           p2 - (Ca * Cim * Ra_micro * (Rv + Rv_micro)),
           q0 - (Ra + Ra_micro + Rv + Rv_micro),
           q1 - (Ra * Ca * (Ra_micro + Rv + Rv_micro) + Cim * (Ra + Ra_micro) * (Rv + Rv_micro)),
           q2 - (Ca * Cim * Ra * Ra_micro * (Rv + Rv_micro)),
           b1 - (Cim * (Rv + Rv_micro))]
    assert np.max(np.abs(res)) < 1e-5, 'SV coronary constants inconsistent'

    # export constants
    return {'Ra1': Ra, 'Ra2': Ra_micro, 'Ca': Ca, 'Cc': Cim, 'Rv1': Rv, 'P_v': 0.0}



import pytest
import numpy as np
from svv.domain.core.h_matrix import h_matrix


def test_h_matrix_shape():
    """
    Test that the H matrix and its components have the correct shape.
    """
    n, d, lam = 4, 2, 0.1  # 4 points in 2D, regularization parameter
    size = (n + 1) * (d + 1)  # Shape of the interpolation matrix A
    a_ = np.random.rand(size, size)

    h, j00, j01, j11, a_inv = h_matrix(a_, n, d, lam)

    # H matrix shape should be (d * n, d * n)
    assert h.shape == (d * n, d * n), f"Expected H matrix shape {(d * n, d * n)}, got {h.shape}"

    # j00_ should have shape (n, n)
    assert j00.shape == (n, n), f"Expected j00 shape {(n, n)}, got {j00.shape}"

    # j01_ should have shape (n, d * n)
    assert j01.shape == (n, d * n), f"Expected j01 shape {(n, d * n)}, got {j01.shape}"

    # j11_ should have shape (d * n, d * n)
    assert j11.shape == (d * n, d * n), f"Expected j11 shape {(d * n, d * n)}, got {j11.shape}"

    # a_inv should have the same shape as a_
    assert a_inv.shape == a_.shape, f"Expected a_inv shape {a_.shape}, got {a_inv.shape}"

def test_h_matrix_no_regularization():
    """
    Test H matrix computation when lam = 0 (no regularization).
    """
    n, d, lam = 3, 3, 0.0  # 3 points in 3D, no regularization
    size = (n + 1) * (d + 1)
    a_ = np.random.rand(size, size)
    h, j00, j01, j11, a_inv = h_matrix(a_, n, d, lam)

    # H matrix should equal j11_ when lam = 0
    assert np.allclose(h, j11), "H matrix should equal j11 when lam = 0."


def test_h_matrix_with_regularization():
    """
    Test H matrix computation when lam > 0 (with regularization).
    """
    n, d, lam = 3, 2, 0.1  # 3 points in 2D, regularization parameter
    size = (n + 1) * (d + 1)
    a_ = np.random.rand(size, size)
    h, j00, j01, j11, a_inv = h_matrix(a_, n, d, lam)

    # Compute the expected H matrix with regularization
    j00_inv = np.linalg.inv(np.eye(j00.shape[0]) + lam * j00)
    expected_h = j11 - lam * j01.T @ j00_inv @ j01

    assert np.allclose(h, expected_h), "H matrix computation with regularization is incorrect."


def test_h_matrix_a_inv_consistency():
    """
    Test that the computed inverse (a_inv) is consistent with the original matrix (a_).
    """
    n, d, lam = 3, 2, 0.1  # 3 points in 2D, regularization parameter
    size = (n + 1) * (d + 1)
    a_ = np.random.rand(size, size)

    _, _, _, _, a_inv = h_matrix(a_, n, d, lam)

    # Verify that A @ A_inv is approximately the identity matrix
    identity = np.eye(a_.shape[0])
    assert np.allclose(a_ @ a_inv, identity), "a_inv is not the correct inverse of a_."


def test_h_matrix_positive_semidefinite():
    """
    Test that the computed H matrix is positive semi-definite.
    """
    n, d, lam = 3, 2, 0.1  # 3 points in 2D, regularization parameter
    size = (n + 1) * (d + 1)
    a_ = np.random.rand(size, size)
    a_ = a_ @ a_.T  # Make a_ symmetric positive definite

    h, _, _, _, _ = h_matrix(a_, n, d, lam)

    # Eigenvalues of H should be non-negative
    eigenvalues = np.linalg.eigvalsh(h)
    assert np.all(eigenvalues >= 0), "H matrix is not positive semi-definite."


def test_h_matrix_edge_case():
    """
    Test edge case when n or d is zero.
    """
    a_ = np.array([[1]])  # Minimal valid A matrix
    h, j00, j01, j11, a_inv = h_matrix(a_, n=0, d=1, lam=0.1)

    # H matrix should be empty
    assert h.size == 0, "H matrix should be empty when n=0."

    # j00_, j01_, j11_, and a_inv should have appropriate shapes
    assert j00.size == 0, "j00 should be empty when n=0."
    assert j01.size == 0, "j01 should be empty when n=0."
    assert j11.size == 0, "j11 should be empty when n=0."
    assert a_inv.shape == (1, 1), f"a_inv shape should be (1, 1), got {a_inv.shape}."



import pytest
import numpy as np
from svv.domain.core.n_matrix import n_matrix


def test_n_matrix_basic_2d():
    """
    Test the N matrix for a simple 2D input.
    """
    points = np.array([
        [0.0, 0.0],
        [1.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0]
    ])
    expected_shape = (12, 3)  # (D+1)*N rows, D+1 columns where N=4, D=2
    result = n_matrix(points)

    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"

    # Check some specific values in the matrix
    # First row corresponds to [0, 0, 1] (point with a 1 appended)
    assert np.allclose(result[0], [0.0, 0.0, 1.0])
    # Second row corresponds to [1, 0, 1]
    assert np.allclose(result[1], [1.0, 0.0, 1.0])
    # Fifth row (derivative constraint) should have [1, 0, 0]
    assert np.allclose(result[4], [1.0, 0.0, 0.0])


def test_n_matrix_random_3d():
    """
    Test the N matrix for random 3D points.
    """
    np.random.seed(0)
    points = np.random.rand(5, 3)  # 5 points in 3D
    expected_shape = (20, 4)  # (D+1)*N rows, D+1 columns where N=5, D=3
    result = n_matrix(points)

    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"

    # Check if the first N rows have correct point values with a 1 appended
    for i in range(points.shape[0]):
        assert np.allclose(result[i, :-1], points[i])
        assert result[i, -1] == 1.0, f"Last column in row {i} should be 1."

    # Check derivative constraints rows
    for i in range(points.shape[0]):
        derivative_block = result[points.shape[0] + i * points.shape[1]:points.shape[0] + (i + 1) * points.shape[1], :3]
        assert np.allclose(derivative_block, np.eye(3)), f"Derivative block for point {i} is incorrect."


def test_n_matrix_edge_case_empty_points():
    """
    Test the N matrix for an empty input.
    """
    points = np.empty((0, 3))  # No points in 3D
    expected_shape = (0, 4)  # No rows, (D+1) columns
    result = n_matrix(points)

    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"
    assert result.size == 0, "Result should be an empty matrix."


def test_n_matrix_edge_case_single_point():
    """
    Test the N matrix for a single point.
    """
    points = np.array([[1.0, 2.0, 3.0]])  # Single point in 3D
    expected_shape = (4, 4)  # (D+1)*N rows, D+1 columns where N=1, D=3
    result = n_matrix(points)

    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"

    # Check the first row corresponds to the point with 1 appended
    assert np.allclose(result[0], [1.0, 2.0, 3.0, 1.0])
    # Check the derivative rows
    derivative_block = result[1:, :3]
    assert np.allclose(derivative_block, np.eye(3)), "Derivative block is incorrect for a single point."


def test_n_matrix_high_dimensional_points():
    """
    Test the N matrix for points in a high-dimensional space (5D).
    """
    points = np.random.rand(3, 5)  # 3 points in 5D
    expected_shape = (18, 6)  # (D+1)*N rows, D+1 columns where N=3, D=5
    result = n_matrix(points)

    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"

    # Check the first N rows have correct point values with a 1 appended
    for i in range(points.shape[0]):
        assert np.allclose(result[i, :-1], points[i])
        assert result[i, -1] == 1.0, f"Last column in row {i} should be 1."

    # Check derivative constraints rows
    for i in range(points.shape[0]):
        derivative_block = result[points.shape[0] + i * points.shape[1]:points.shape[0] + (i + 1) * points.shape[1], :5]
        assert np.allclose(derivative_block, np.eye(5)), f"Derivative block for point {i} is incorrect."



import numpy


class TreeMap(dict):
    """
    The TreeMap class defines a mapping between vessels
    in the tree and their upstream and downstream vessels.
    """
    def __new__(cls, *args, **kwargs):
        data = super().__new__(cls, *args, **kwargs)
        return data


class TreeParameters(object):
    def __init__(self):
        """
        The TreeParameters class defines the parameters
        that are used to generate a synthetic vascular
        tree.
        """
        self.kinematic_viscosity = 3.6e-2
        self.fluid_density = 1.06
        self.murray_exponent = 3.0
        self.radius_exponent = 2.0
        self.length_exponent = 1.0
        self.terminal_pressure = 60.0 * 1333.22
        self.root_pressure = 100.0 * 1333.22
        self.terminal_flow = 0.125/2000
        self.root_flow = None
        self.max_nonconvex_count = 100

    def __str__(self):
        output = "Tree Parameters:\n"
        output += "----------------\n"
        output += "Kinematic Viscosity: {}\n".format(self.kinematic_viscosity)
        output += "Fluid Density: {}\n".format(self.fluid_density)
        output += "Murray Exponent: {}\n".format(self.murray_exponent)
        output += "Radius Exponent: {}\n".format(self.radius_exponent)
        output += "Length Exponent: {}\n".format(self.length_exponent)
        output += "Terminal Pressure: {}\n".format(self.terminal_pressure)
        output += "Root Pressure: {}\n".format(self.root_pressure)
        output += "Terminal Flow: {}\n".format(self.terminal_flow)
        output += "Root Flow: {}".format(self.root_flow)
        return output

    def __repr__(self):
        output = "Tree Parameters:\n"
        output += "----------------\n"
        output += "Kinematic Viscosity: {}\n".format(self.kinematic_viscosity)
        output += "Fluid Density: {}\n".format(self.fluid_density)
        output += "Murray Exponent: {}\n".format(self.murray_exponent)
        output += "Radius Exponent: {}\n".format(self.radius_exponent)
        output += "Length Exponent: {}\n".format(self.length_exponent)
        output += "Terminal Pressure: {}\n".format(self.terminal_pressure)
        output += "Root Pressure: {}\n".format(self.root_pressure)
        output += "Terminal Flow: {}\n".format(self.terminal_flow)
        output += "Root Flow: {}".format(self.root_flow)
        return output

    def set(self, parameter, value):
        """
        Set a parameter of the tree.
        """
        if parameter == 'kinematic_viscosity':
            self.kinematic_viscosity = value
        elif parameter == 'fluid_density':
            self.fluid_density = value
        elif parameter == 'murray_exponent':
            self.murray_exponent = value
        elif parameter == 'radius_exponent':
            self.radius_exponent = value
        elif parameter == 'length_exponent':
            self.length_exponent = value
        elif parameter == 'terminal_pressure':
            self.terminal_pressure = value
        elif parameter == 'root_pressure':
            self.root_pressure = value
        elif parameter == 'terminal_flow':
            self.terminal_flow = value
        elif parameter == 'root_flow':
            self.root_flow = value
        else:
            raise ValueError("Invalid parameter: {}.".format(parameter))
        return None


class TreeData(numpy.ndarray):
    """
    TreeData is a subclass of the numpy.ndarray class
    that is used to store synthetic vascular data.
    """
    def __new__(cls, *args, **kwargs):
        if len(args) == 0:
            shape = (1, 31)
            data = super().__new__(cls, shape, **kwargs)
            data[:] = numpy.nan
            return data
        elif isinstance(args[0], numpy.ndarray):
            return args[0].view(cls)
        else:
            data = super().__new__(cls, *args, **kwargs)
            data[:] = numpy.nan
            return data

    @classmethod
    def from_array(cls, arr):
        if isinstance(arr, numpy.ndarray):
            return arr.view(cls)
        else:
            return None

    def __array_finalize__(self, obj):
        if obj is None:
            return None

    def __proximal__(self, *args):
        if len(args) == 0:
            return self[:, 0:3].copy()
        else:
            return self[args[0], 0:3].copy()

    def __distal__(self, *args):
        if len(args) == 0:
            return self[:, 3:6].copy()
        else:
            return self[args[0], 3:6].copy()

    def __u_basis__(self, *args):
        if len(args) == 0:
            return self[:, 6:9].copy()
        else:
            return self[args[0], 6:9].copy()

    def __v_basis__(self, *args):
        if len(args) == 0:
            return self[:, 9:12].copy()
        else:
            return self[args[0], 9:12].copy()

    def __w_basis__(self, *args):
        if len(args) == 0:
            return self[:, 12:15].copy()
        else:
            return self[args[0], 12:15].copy()

    def __children__(self, *args):
        if len(args) == 0:
            return self[:, 15:17].copy()
        else:
            return self[args[0], 15:17].copy()

    def __parent__(self, *args):
        if len(args) == 0:
            return self[:, 17].copy()
        else:
            return self[args[0], 17].copy()

    def __proximal_node__(self, *args):
        if len(args) == 0:
            return self[:, 18].copy()
        else:
            return self[args[0], 18].copy()

    def __distal_node__(self, *args):
        if len(args) == 0:
            return self[:, 19].copy()
        else:
            return self[args[0], 19].copy()

    def __length__(self, *args):
        if len(args) == 0:
            return self[:, 20].copy()
        else:
            return self[args[0], 20].copy()

    def __radius__(self, *args):
        if len(args) == 0:
            return self[:, 21].copy()
        else:
            return self[args[0], 21].copy()

    def __flow__(self, *args):
        if len(args) == 0:
            return self[:, 22].copy()
        else:
            return self[args[0], 22].copy()

    def __left_bifurcation__(self, *args):
        if len(args) == 0:
            return self[:, 23].copy()
        else:
            return self[args[0], 23].copy()

    def __right_bifurcation__(self, *args):
        if len(args) == 0:
            return self[:, 24].copy()
        else:
            return self[args[0], 24].copy()

    def __reduced_resistance__(self, *args):
        if len(args) == 0:
            return self[:, 25].copy()
        else:
            return self[args[0], 25].copy()

    def __depth__(self, *args):
        if len(args) == 0:
            return self[:, 26].copy()
        else:
            return self[args[0], 26].copy()

    def __reduced_length__(self, *args):
        if len(args) == 0:
            return self[:, 27].copy()
        else:
            return self[args[0], 27].copy()

    def __radius_scaling__(self, *args):
        if len(args) == 0:
            return self[:, 28].copy()
        else:
            return self[args[0], 28].copy()

    def get(self, label, *args):
        if label == 'proximal':
            return self.__proximal__(*args)
        elif label == 'distal':
            return self.__distal__(*args)
        elif label == 'u_basis':
            return self.__u_basis__(*args)
        elif label == 'v_basis':
            return self.__v_basis__(*args)
        elif label == 'w_basis':
            return self.__w_basis__(*args)
        elif label == 'children':
            return self.__children__(*args)
        elif label == 'parent':
            return self.__parent__(*args)
        elif label == 'proximal_node':
            return self.__proximal_node__(*args)
        elif label == 'distal_node':
            return self.__distal_node__(*args)
        elif label == 'length':
            return self.__length__(*args)
        elif label == 'radius':
            return self.__radius__(*args)
        elif label == 'flow':
            return self.__flow__(*args)
        elif label == 'left_bifurcation':
            return self.__left_bifurcation__(*args)
        elif label == 'right_bifurcation':
            return self.__right_bifurcation__(*args)
        elif label == 'reduced_resistance':
            return self.__reduced_resistance__(*args)
        elif label == 'depth':
            return self.__depth__(*args)
        elif label == 'reduced_length':
            return self.__reduced_length__(*args)
        elif label == 'radius_scaling':
            return self.__radius_scaling__(*args)
        else:
            raise ValueError("Invalid label.")



import numpy as np
from ..core.a_matrix import a_matrix
from ..core.h_matrix import h_matrix
from .cost import cost
from .coordinate_system import cart2sph


class Kernel:
    r"""
    Kernel class for variational implicit point set surfaces.

    This class defines the minimization kernel for the implicit domain object used in the construction of variational implicit point set surfaces, as described in Huang et al. [1]_.

    The `Kernel` class sets up the optimization problem for energy minimization of an implicit surface defined by unstructured point cloud data. It computes the necessary matrices and provides methods to evaluate the cost function, its gradient, and Hessian for optimization algorithms.

    Parameters
    ----------
    points : ndarray of shape (N, D)
        The input point cloud data defining the surface. Each row corresponds to a point in D-dimensional space.

    rbf_degree : int, optional
        The degree of the radial basis function used in constructing the interpolation matrix. Default is 3.

    lam : float, optional
        The regularization parameter (lambda) controlling the smoothness of the resulting surface. Default is 0.

    Attributes
    ----------
    n : int
        The number of points in the point cloud (N).

    d : int
        The dimensionality of the embedding space (D).

    a_ : ndarray of shape ((D+1)*N, (D+1)*N)
        The full interpolation matrix \( \mathbf{A} \) constructed using radial basis functions and polynomial terms.

    h_ : ndarray of shape (D*N, D*N)
        The \( \mathbf{H} \) matrix used in the energy minimization problem.

    a_inv : ndarray of shape ((D+1)*N, (D+1)*N)
        The inverse of the interpolation matrix \( \mathbf{A}^{-1} \).

    lam : float
        The regularization parameter controlling the smoothness of the surface.

    x0 : list of ndarray
        Initial values for the optimization variables.

    h0 : list of ndarray
        Corresponding \( \mathbf{H} \) matrices for different regularization parameters.

    Methods
    -------
    set_initial_values(*args, **kwargs)
        Set the initial values and parameters for the optimization problem.

    get_bounds()
        Get the bounds for the optimization variables.

    eval(x)
        Evaluate the cost function at a given point x.

    gradient(x)
        Evaluate the gradient of the cost function at a given point x.

    hessian(x)
        Evaluate the Hessian of the cost function at a given point x.

    Notes
    -----
    **Variational Implicit Point Set Surfaces**

    The variational implicit point set surfaces method aims to reconstruct smooth surfaces from point cloud data by formulating the problem as an energy minimization task. The key idea is to find an implicit function that minimizes an energy functional while interpolating or approximating the given data points. This approach leads to surfaces that are smooth and can handle noisy data gracefully.

    **Optimization Problem Setup**

    The Kernel class prepares the necessary components for setting up the optimization problem:

    - Constructs the interpolation matrix \( \mathbf{A} \) using radial basis functions.
    - Computes the \( \mathbf{H} \) matrix used in the bending energy minimization.
    - Provides methods to set initial values, define variable bounds, and evaluate the cost function and its derivatives.

    References
    ----------
    .. [1] Huang, J., Zhang, C., Han, H., & Zwicker, M. (2019). Variational implicit point set surfaces. *ACM Transactions on Graphics (TOG)*, 38(4), 1-14. https://doi.org/10.1145/3306346.3322994

    Examples
    --------
    **Example:** Setting up and using the Kernel class for surface reconstruction.

    .. code-block:: python

        import numpy as np
        from your_package.kernel import Kernel

        # Generate sample point cloud data
        points = np.random.rand(100, 3)  # 100 points in 3D space

        # Initialize the Kernel object
        kernel = Kernel(points, rbf_degree=3, lam=0.1)

        # Set initial values for optimization
        kernel.set_initial_values()

        # Get the bounds for the optimization variables
        bounds = kernel.get_bounds()

        # Define an optimization routine (e.g., using scipy.optimize)
        from scipy.optimize import minimize

        # Use the first initial guess and cost function
        x0 = kernel.x0[0]
        cost_function = kernel.__costs__[0]

        # Perform optimization
        result = minimize(cost_function, x0, bounds=bounds)

        # Extract optimized parameters
        optimized_params = result.x

        # Evaluate the cost at the optimized parameters
        final_cost = kernel.eval(optimized_params)
        print("Final cost:", final_cost)

    See Also
    --------
    a_matrix : Function to compute the full interpolation matrix \( \mathbf{A} \).
    h_matrix : Function to compute the \( \mathbf{H} \) matrix for bending energy minimization.
    cost : Function to compute the cost function value.
    cart2sph : Function to convert Cartesian coordinates to spherical coordinates.
    sph2cart : Function to convert spherical coordinates to Cartesian coordinates.

    """

    def __init__(self, points, rbf_degree=3, lam=0):
        """
        Initialize the Kernel object.

        Parameters
        ----------
        points : ndarray of shape (N, D)
            The input point cloud data defining the surface.

        rbf_degree : int, optional
            The degree of the radial basis function used in constructing the interpolation matrix. Default is 3.

        lam : float, optional
            The regularization parameter (lambda) controlling the smoothness of the resulting surface. Default is 0.

        """
        self.n = points.shape[0]
        self.d = points.shape[1]
        self.a_ = a_matrix(points, rbf_degree)
        self.h_, self.j00_, self.j01_, self.j11_, self.a_inv = h_matrix(self.a_, self.n, self.d, lam)
        self.__cost__ = lambda x: cost(x, self.h_, self.n, self.d)
        self.__costs__ = None
        self.__grad__ = None
        self.__hess__ = None
        self.x0 = None
        self.lam = lam
        self.h0 = None

    def set_initial_values(self, *args, **kwargs):
        r"""
        Set initial values and parameters for the optimization problem.

        This method prepares the initial guesses for the optimization variables and, if necessary, adjusts the regularization parameters.

        Parameters
        ----------
        *args : optional
            - If provided, the first positional argument should be an ndarray of normals corresponding to the points.

        **kwargs : optional
            - lambda_parameters : list or tuple of shape (3,), optional
                Parameters to linearly vary the thin-plate relaxation parameter (lambda). The list should contain:
                [minimum value, maximum value, number of values to generate].

                Default is [0.001, 1, 5].

        Returns
        -------
        None

        Notes
        -----
        **Usage Scenarios:**

        - **Using Normals as Initial Values:**

          If normals are provided as positional arguments, they are normalized and converted to spherical coordinates to serve as initial guesses.

        - **Varying Lambda Parameters:**

          If no normals are provided, the method will generate initial values by varying the regularization parameter lambda within the specified range.

        Examples
        --------
        **Example 1:** Using normals to set initial values.

        .. code-block:: python

            normals = np.random.rand(100, 3) - 0.5  # Random normals
            kernel.set_initial_values(normals)

        **Example 2:** Specifying lambda parameters.

        .. code-block:: python

            kernel.set_initial_values(lambda_parameters=[0.01, 0.5, 10])

        """
        if len(args) > 0:
            normals = args[0]
            normals = normals / np.linalg.norm(normals, axis=1).reshape(-1, 1)
            spherical_coordinates = cart2sph(normals)
            x0 = spherical_coordinates[:, 1:]
            h0 = [self.h_]
            x0 = [x0.flatten()]
            funcs = [lambda x: cost(x, self.h_, self.n, self.d)]
        else:
            lam_params = kwargs.get("lambda_parameters", [0.001, 1, 5])
            lams = np.linspace(self.lam + lam_params[0], self.lam + lam_params[1], lam_params[2])
            h0 = []
            x0 = []
            funcs = []
            for lam in lams:
                h_, _, _, _, _ = h_matrix(self.a_, self.n, self.d, lam)
                h0.append(h_)
                eigen_values, eigen_vectors = np.linalg.eig(h_)
                eh = np.argmin(eigen_values)
                x0_init = eigen_vectors[:, eh].real
                x0_init = x0_init.reshape(self.n, self.d)
                x0_init = x0_init / np.linalg.norm(x0_init, axis=1).reshape(-1, 1)
                x0_init = cart2sph(x0_init)
                x0_init = x0_init[:, 1:]
                x0_init = x0_init.flatten()
                x0.append(x0_init)
                func = lambda x: cost(x, h_, self.n, self.d)
                funcs.append(func)
        self.x0 = x0
        self.h0 = h0
        self.__costs__ = funcs
        return None

    def get_bounds(self):
        r"""
        Get the bounds for the optimization variables.

        Returns
        -------
        bounds : tuple of two lists
            A tuple containing two lists:
            - Lower bounds for each optimization variable.
            - Upper bounds for each optimization variable.

        Notes
        -----
        The bounds are set based on the ranges of the angular components in spherical coordinates:

        - For angles \( \phi_i \) where \( i < d - 1 \), the bounds are \( [0, \pi] \).
        - For the azimuthal angle \( \phi_{d-1} \), the bounds are \( [0, 2\pi] \).

        Examples
        --------
        .. code-block:: python

            bounds = kernel.get_bounds()
            # Use bounds in an optimization routine
            result = minimize(cost_function, x0, bounds=bounds)

        """
        bounds = []
        lb = []
        ub = []
        for i in range(self.n):
            for j in range(self.d - 1):
                if j < self.d - 2:
                    lb.append(0)
                    ub.append(np.pi)
                else:
                    lb.append(0)
                    ub.append(2 * np.pi)
        bounds.append(lb)
        bounds.append(ub)
        return tuple(bounds)

    def eval(self, x):
        r"""
        Evaluate the cost function at a given point.

        Parameters
        ----------
        x : ndarray
            The optimization variables (angular components) at which to evaluate the cost function.

        Returns
        -------
        cost_value : float
            The value of the cost function at the given point.

        Examples
        --------
        .. code-block:: python

            x = kernel.x0[0]  # Initial guess
            cost_value = kernel.eval(x)
            print("Cost at x:", cost_value)

        """
        return self.__cost__(x)

    def gradient(self, x):
        r"""
        Evaluate the gradient of the cost function at a given point.

        **Note:** This method is currently a placeholder and needs to be implemented.

        Parameters
        ----------
        x : ndarray
            The optimization variables at which to evaluate the gradient.

        Returns
        -------
        grad : ndarray
            The gradient vector of the cost function at the given point.

        """
        if self.__grad__ is not None:
            return self.__grad__(x)
        else:
            raise NotImplementedError("Gradient function is not implemented.")

    def hessian(self, x):
        r"""
        Evaluate the Hessian of the cost function at a given point.

        **Note:** This method is currently a placeholder and needs to be implemented.

        Parameters
        ----------
        x : ndarray
            The optimization variables at which to evaluate the Hessian.

        Returns
        -------
        hess : ndarray
            The Hessian matrix of the cost function at the given point.

        """
        return self.__hess__(x)



# cython: language_level=3
# distutils: language=c++
import numpy as np
cimport numpy as cnp
import cython
#from cython.parallel import prange, parallel
from libc.math cimport pow, pi
from libcpp.vector cimport vector

cnp.import_array()

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def update_resistance(double[:,:] data, long long[:,:] idx, double gamma, double nu):
    cdef double lr, lbif, rbif
    cdef Py_ssize_t size = data.shape[0]
    cdef vector[long long] vessels = range(size)
    cdef vector[long long] tmp_vessels
    cdef long long i
    max_depth = max(data[:, 26])
    while vessels.size() > 0:
        tmp_vessels.clear()
        for ii in range(vessels.size()):
            if data[vessels[ii], 26] != max_depth:
                tmp_vessels.push_back(vessels[ii])
            else:
                i = vessels[ii]
                if np.isnan(data[i, 15:17]).all():
                    data[i, 25] = (8 * nu / pi) * data[i, 20]
                    data[i, 27] = 0.0
                elif np.isnan(data[i, 15]):
                    right = idx[i, 1]
                    data[i, 25] = (8 * nu / pi) * data[i, 20] + data[right, 25]
                    data[i, 23] = 0.0
                    data[i, 24] = 1.0
                    data[i, 27] = data[right, 20] + data[right, 27]
                elif np.isnan(data[i, 16]):
                    left = idx[i, 0]
                    data[i, 25] = (8 * nu / pi) * data[i, 20] + data[left, 25]
                    data[i, 23] = 1.0
                    data[i, 24] = 0.0
                    data[i, 27] = data[left, 20] + data[left, 27]
                else:
                    left = idx[i, 0]
                    right = idx[i, 1]
                    lr = pow(((data[left, 22]*data[left, 25])/(data[right, 22]*data[right, 25])), 0.25)
                    lbif = pow((1.0 + pow(lr, -gamma)), -1.0/gamma)
                    rbif = pow((1.0 + pow(lr, gamma)), -1.0/gamma)
                    data[i, 25] = (8 * nu / pi) * data[i, 20] + pow(((pow(lbif, 4.0)/data[left, 25]) +
                                                                     (pow(rbif, 4.0)/data[right, 25])), -1.0)
                    data[i, 23] = lbif
                    data[i, 24] = rbif
                    data[i, 27] = pow(lbif, 2.0) * (data[left, 20] + data[left, 27]) + pow(rbif, 2.0) * (data[right, 20] + data[right, 27])
        vessels = tmp_vessels


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def update_radii(double[:,:] data, long long[:,:] idx, double root_pressure, double terminal_pressure):
    cdef vector[long long] left_vessels
    cdef vector[long long] right_vessels
    cdef vector[long long] tmp_left_vessels
    cdef vector[long long] tmp_right_vessels
    cdef long long i
    data[0, 21] = pow(((data[0, 25] * data[0, 22])/(root_pressure - terminal_pressure)), 0.25)
    if ~np.isnan(data[0, 15]):
        left_vessels.push_back(idx[0, 0])
    if ~np.isnan(data[0, 16]):
        right_vessels.push_back(idx[0, 1])
    while left_vessels.size() > 0 or right_vessels.size() > 0:
        tmp_left_vessels.clear()
        tmp_right_vessels.clear()
        for i in left_vessels:
            data[i, 21] = data[idx[i, 3], 23]*data[idx[i, 3], 21]
            if ~np.isnan(data[i, 15]):
                tmp_left_vessels.push_back(idx[i, 0])
            if ~np.isnan(data[i, 16]):
                tmp_right_vessels.push_back(idx[i, 1])
        for i in right_vessels:
            data[i, 21] = data[idx[i, 3], 24] * data[idx[i, 3], 21]
            if ~np.isnan(data[i, 15]):
                tmp_left_vessels.push_back(idx[i, 0])
            if ~np.isnan(data[i, 16]):
                tmp_right_vessels.push_back(idx[i, 1])
        left_vessels = tmp_left_vessels
        right_vessels = tmp_right_vessels




import numpy as np
import os
import uuid
import pyvista
from time import perf_counter
from tqdm import trange
from copy import deepcopy
from types import MappingProxyType
import itertools
from svv.tree.data.data import TreeData, TreeParameters, TreeMap
from svv.tree.branch.root import set_root
from svv.visualize.tree.show import show
from svv.tree.branch.bifurcation import add_vessel, check_tree
from svv.tree.export.export_solid import build_watertight_solid, build_merged_solid
from svv.tree.export.export_centerlines import build_centerlines
from svv.tree.utils.TreeManager import KDTreeManager, HNSWTree, USearchTree
from svv.tree.utils.c_extend import build_c_vessel_map
from collections import ChainMap


class Tree(object):
    def __init__(self):
        """
        The Tree class defines a branching tree structure
        that is used to abstract the physical representation
        of the generated vascular network.
        """
        self.data = TreeData()
        self.parameters = TreeParameters()
        self.vessel_map = TreeMap()
        #self.c_vessel_map = None
        self.physical_clearance = 0.0
        self.domain = None
        self.convex = None
        self.random_seed = None
        self.characteristic_length = None
        self.domain_clearance = None
        self.n_terminals = 0
        self.nonconvex_count = 0
        self.clamped_root = False
        self.rtree = None
        self.preallocation_step = int(4e6)
        self.preallocate = TreeData((self.preallocation_step, 31))
        self.segment_count = 0
        #self.preallocate_connectivity = (np.ones((self.preallocation_step, 3)) * -1).astype(int)
        self.preallocate_midpoints = np.zeros((self.preallocation_step, 3))
        self.times = {'vessels':[],
                      'local_optimization':[],
                      'collision':[],
                      'collision_1':[],
                      'collision_2':[],
                      'get_points':[],
                      'get_points_0':[],
                      'get_points_1':[],
                      'get_points_2':[],
                      'get_points_3':[],
                      'chunk_1':[],
                      'chunk_2':[],
                      'chunk_3':[],
                      'chunk_3_0':[],
                      'chunk_3_1':[],
                      'chunk_3_2':[],
                      'chunk_3_3':[],
                      'chunk_3_4':[],
                      'chunk_3_4_alt':[],
                      'chunk_3_5':[],
                      'chunk_3_6':[],
                      'chunk_3_7':[],
                      'chunk_4':[],
                      'chunk_4_0':[],
                      'chunk_4_1':[],
                      'chunk_4_2':[],
                      'chunk_4_3':[],
                      'chunk_5':[],
                      'all':[]}

    def set_domain(self, domain, convexity_tolerance=1e-2):
        """
        Set the domain for the tree. The domain should be an implicit
        function of the form f(X) = scalar, where X is a cartesian
        coordinate in space.
        """
        self.domain = domain
        self.characteristic_length = domain.characteristic_length
        self.convex = np.isclose(domain.convexity, 1.0, atol=convexity_tolerance)
        if self.physical_clearance > 0.0:
            patch = domain.patches[0]
            pt = (patch.points[0, :] + patch.normals[0, :] * self.physical_clearance).reshape(1, patch.points.shape[1])
            value = domain(pt)
            self.domain_clearance = abs(value).flatten()[0]
        else:
            self.domain_clearance = 0.0
        if self.domain.mesh is not None:
            self.domain.mesh.cell_data['probability'] = self.domain.mesh.cell_data['Normalized_Volume'].copy()
        return None

    def set_root(self, *args, **kwargs):
        """
        Set the root point of the tree.
        """
        if len(args) == 0:
            pass
        elif len(args) == 1:
            kwargs['start'] = args[0]
        elif len(args) == 2:
            kwargs['start'] = args[0]
            kwargs['direction'] = args[1]
            self.clamped_root = True
        else:
            raise ValueError("Too many arguments.")
        inplace = kwargs.pop('inplace', True)
        if not isinstance(kwargs.get('direction', None), type(None)):
            self.clamped_root = True
        if self.physical_clearance > 0.0:
            kwargs['interior_range'] = [-1.0, 0.0-self.domain_clearance]
        root, root_map = set_root(self, **kwargs)
        if inplace:
            self.data = root
            self.preallocate[0, :] = root
            #self.data_copy = self.preallocate[:3, :]
            self.connectivity = np.nan_to_num(root[:, 15:18], nan=-1.0).astype(int).reshape(1, 3)
            #self.preallocate_connectivity[0, :] = self.connectivity
            #self.connectivity_copy = self.connectivity.copy()
            #self.connectivity_copy = self.preallocate_connectivity[:3, :]
            self.preallocate_midpoints[0, :] = (root[:, 0:3] + root[:, 3:6]) / 2
            self.midpoints = self.preallocate_midpoints[0, :]
            #self.midpoints_copy = self.preallocate_midpoints[:1, :]
            self.vessel_map.update(root_map)
            #self.c_vessel_map = build_c_vessel_map(root_map)
            self.vessel_map_copy = deepcopy(self.vessel_map)
            self.n_terminals = 1
            self.kdtm = KDTreeManager(((root[:, 0:3] + root[:, 3:6]) / 2).reshape(1, 3))
            self.hnsw_tree = USearchTree(((root[:, 0:3] + root[:, 3:6]) / 2).reshape(1, 3).astype(np.float32))
            self.hnsw_tree_id = id(self.hnsw_tree)
            self.probability = np.array(self.domain.mesh.cell_data['probability'])
            self.max_distal_node = 1
            self.tree_scale = np.pi * root[0, 21]**self.parameters.radius_exponent*root[0, 20]**self.parameters.length_exponent
            self.segment_count = 1
            #self.rtree = RTree()
            #self.rtree.insert(((root[:, 0:3] + root[:, 3:6]) / 2).reshape(1, 3))
        else:
            connectivity = np.nan_to_num(root[:, 15:18], nan=-1.0).astype(int).reshape(1, 3)
            kdtm = KDTreeManager(((root[:, 0:3] + root[:, 3:6]) / 2).reshape(1, 3))
            hnsw_tree = USearchTree(((root[:, 0:3] + root[:, 3:6]) / 2).reshape(1, 3).astype(np.float32))
            hnsw_tree_id = id(hnsw_tree)
            probability = np.array(self.domain.mesh.cell_data['probability'])
            tree_scale = np.pi * root[0, 21] ** self.parameters.radius_exponent * root[
                0, 20] ** self.parameters.length_exponent
            return root, root_map, connectivity, kdtm, hnsw_tree, hnsw_tree_id, probability, tree_scale

    def add(self, inplace=True, **kwargs):
        all_start = perf_counter()
        decay_probability = kwargs.pop('decay_probability', 0.9)
        new_data, added_vessels, new_vessel_map, history, lines, nonconvex_outside, new_inds, mesh_cell, connectivity, change_i, change_j, new_tmp_data, old_tmp_data = add_vessel(self, **kwargs)
        start = perf_counter()
        #native_volume = np.sum(np.pi*new_data[:, 21]**2.0*new_data[:, 20])
        #assert np.isclose(native_volume,self.new_tree_scale), "native {} cost {} volumes do not match ".format(native_volume, self.new_tree_scale)
        #results = tuple([new_data, added_vessels, new_vessel_map, history, lines])
        #check = check_tree(self, results)
        #if not check[0]:
        #    print("Tree is not valid.")
        #    return check[1], check[2]
        #start = perf_counter()
        if not self.convex:
            if nonconvex_outside:
                self.nonconvex_count = 0
            else:
                self.nonconvex_count += 1
            if self.nonconvex_count > self.parameters.max_nonconvex_count:
                self.convex = True
        if inplace:
            #start_5 = perf_counter()
            #_data = TreeData(new_data.shape)
            #_data[:, :] = new_data[:, :]
            #self.data = _data
            #end_5 = perf_counter()
            #self.times['chunk_5'].append(end_5-start_5)
            start_chunk_4_0 = perf_counter()
            #new_data = np.vstack([self.data, added_vessels[0], added_vessels[1]])
            self.preallocate[self.segment_count,:] = added_vessels[0]
            self.preallocate[self.segment_count+1,:] = added_vessels[1]
            end_chunk_4_0 = perf_counter()
            self.times['chunk_4_0'].append(end_chunk_4_0 - start_chunk_4_0)
            start_chunk_4_1 = perf_counter()
            #new_data[change_i, change_j] = np.array(new_tmp_data)
            change_i = np.array(change_i, dtype=int)
            change_j = np.array(change_j, dtype=int)
            #if change_i.shape[0] > 10000:
            #    #print('parallel')
            #    parallel_scatter_update(self.preallocate, change_i, change_j, np.array(new_tmp_data))
            #else:
            self.preallocate[change_i, change_j] = np.array(new_tmp_data)
            end_chunk_4_1 = perf_counter()
            self.times['chunk_4_1'].append(end_chunk_4_1 - start_chunk_4_1)
            start_chunk_4_2 = perf_counter()
            #self.data = TreeData.from_array(new_data)
            self.data = self.preallocate[:self.segment_count+2,:]
            new_data = self.data
            end_chunk_4_2 = perf_counter()
            self.times['chunk_4_2'].append(end_chunk_4_2 - start_chunk_4_2)
            #np.copyto(self.preallocate[:new_data.shape[0], :], new_data)
            #self.data_copy = self.preallocate[:(new_data.shape[0] + 2), :]
            #self.vessel_map_copy = ChainMap(new_vessel_map, self.vessel_map)
            start_chunk_4_3 = perf_counter()
            for key in new_vessel_map.keys():
                if key in self.vessel_map.keys():
                    #print("new_vessel_map: {}".format(new_vessel_map))
                    #print("vessel_map: {}".format(self.vessel_map))
                    self.vessel_map[key]['upstream'].extend(new_vessel_map[key]['upstream'])
                    #_, counts = np.unique(self.vessel_map[key]['upstream'], return_counts=True)
                    #assert np.all(counts == 1), "Fail in appending upstream idxs.\nkey: {}\nprevious: {}\nextension: {}".format(key, self.vessel_map[key]['upstream'],
                    #                                                                                                   new_vessel_map[key]['upstream'])
                    self.vessel_map[key]['downstream'].extend(new_vessel_map[key]['downstream'])
                    #_, counts = np.unique(self.vessel_map[key]['downstream'], return_counts=True)
                    #assert np.all(counts == 1), "Fail in appending downstream idxs"
                else:
                    self.vessel_map[key] = deepcopy(new_vessel_map[key])
            end_chunk_4_3 = perf_counter()
            self.times['chunk_4_3'].append(end_chunk_4_3 - start_chunk_4_3)
            #self.vessel_map = ChainMap(new_vessel_map, self.vessel_map)
            self.n_terminals += 1
            self.segment_count += 2
            self.connectivity = connectivity #self.connectivity_copy.view()
            #self.connectivity_copy = self.preallocate_connectivity[:(self.connectivity_copy.shape[0] + 2), :]
            self.max_distal_node += 2
            if mesh_cell >= 0:
                self.probability[mesh_cell] *= decay_probability
                self.probability = self.probability / self.probability.sum()
                self.domain.cumulative_probability = np.cumsum(self.probability)
                #self.domain.mesh.cell_data['probability'][mesh_cell] *= decay_probability
                #self.domain.mesh.cell_data['probability'] = self.domain.mesh.cell_data['probability'] / \
                #                                            self.domain.mesh.cell_data['probability'].sum()
                #self.domain.cumulative_probability = np.cumsum(self.domain.mesh.cell_data['probability'])
            #self.kdtm.start_update(((new_data[:, 0:3] + new_data[:, 3:6]) / 2))
            #self.rtree.replace(new_inds[0], ((new_data[new_inds[0], 0:3] + new_data[new_inds[0], 3:6]) / 2).flatten().tolist())
            #self.rtree.rtree_index.insert(new_inds[1], ((new_data[new_inds[1], 0:3] + new_data[new_inds[1], 3:6]) / 2).flatten().tolist())
            #self.rtree.points.append(((new_data[new_inds[1], 0:3] + new_data[new_inds[1], 3:6]) / 2).flatten().tolist())
            #self.rtree.rtree_index.insert(new_inds[2], ((new_data[new_inds[2], 0:3] + new_data[new_inds[2], 3:6]) / 2).flatten().tolist())
            #self.rtree.points.append(((new_data[new_inds[2], 0:3] + new_data[new_inds[2], 3:6]) / 2).flatten().tolist())
            #self.kdtm.wait_for_update()
            #self.kdtm.swap_trees()
            #if self.n_terminals % 10000 == 0:
            #    self.hnsw_tree = USearchTree(((new_data[:, 0:3] + new_data[:, 3:6])/2).astype(np.float32))
            #else:
            self.hnsw_tree.replace(((new_data[new_inds[0], 0:3] + new_data[new_inds[0], 3:6]) / 2).reshape(1, 3).astype(np.float32), np.array([new_inds[0]]))
            self.hnsw_tree.add_items(((new_data[new_inds[1], 0:3] + new_data[new_inds[1], 3:6]) / 2).reshape(1, 3).astype(np.float32), np.array([new_inds[1]]))
            self.hnsw_tree.add_items(((new_data[new_inds[2], 0:3] + new_data[new_inds[2], 3:6]) / 2).reshape(1, 3).astype(np.float32), np.array([new_inds[2]]))
            #self.preallocate_midpoints[new_inds, :] = (new_data[new_inds, 0:3] + new_data[new_inds, 3:6])/2
            #self.midpoints_copy = self.preallocate_midpoints[:new_data.shape[0], :]
            self.preallocate_midpoints[:self.segment_count, :] = (new_data[:, 0:3] + new_data[:, 3:6])/2
            self.tree_scale = self.new_tree_scale
            end = perf_counter()
            self.times['chunk_4'].append(end - start)
            self.times['all'].append(end - all_start)
            return None
        else:
            end = perf_counter()
            self.times['chunk_4'].append(end - start)
            self.times['all'].append(end - all_start)
            #return new_data, added_vessels, new_vessel_map, history, lines, nonconvex_outside, new_inds, mesh_cell
            return change_i, change_j, new_tmp_data, old_tmp_data, new_vessel_map, connectivity, new_inds, mesh_cell, added_vessels

    def n_add(self, n, **kwargs):
        for i in trange(n, desc='Adding vessels', unit='vessel', leave=False):
            self.add(**kwargs)
        return None
    def prune(self):
        pass

    def show(self, **kwargs):
        return show(self, **kwargs)

    def save(self):
        pass

    def export_solid(self, outdir=None, shell_thickness=0.0, watertight=False, **kwargs):
        if isinstance(outdir, type(None)):
            if not os.path.exists(os.getcwd() + os.sep + '3d_tmp'):
                outdir = os.getcwd() + os.sep + '3d_tmp'
            else:
                outdir = os.getcwd() + os.sep + '3d_tmp_' + str(uuid.uuid4())
        else:
            if os.path.exists(outdir):
                raise ValueError("Output directory already exists.")
        if not watertight:
            model = build_merged_solid(self)

        else:
            model = build_watertight_solid(self)
        return model


    def export_centerlines(self, outdir=None, **kwargs):
        centerlines, polys = build_centerlines(self)
        return centerlines, polys


    def export_gcode(self):
        pass

    def export_centerline(self):
        pass




import numpy as np


def cart2sph(cart):
    r"""
    Convert Cartesian coordinates to spherical coordinates.

    This function transforms a set of points in d-dimensional Cartesian coordinates to d-dimensional
    spherical coordinates. In spherical coordinates, the first element is the radial distance (norm), and
    the subsequent elements represent angular coordinates. More information on d-dimensional spherical
    coordinates can be found in [1]_.

    - **Coordinate Conversion**:


    Parameters
    ----------
    cart : ndarray of shape (N, d)
        Cartesian coordinates for N points in d-dimensional space. Each row represents the
        Cartesian coordinates of a point.

    Returns
    -------
    sph : ndarray of shape (N, d)
        Spherical coordinates for the N points. Each row represents the spherical coordinates of a
        point, where the first coordinate is the radial distance (norm), and the remaining coordinates
        are angles in radians.


    Notes
    -----
    The conversion from Cartesian to spherical coordinates in n-dimensional space is performed using
    the following formulas [2]_:

    For a point :math:`\mathbf{x} = (x_1, x_2, x_3, \dots, x_{d-1}, x_d)`:

    - **Radial distance** \( :math:`\mathbf{r}` \):

      .. math::

          r = \sqrt{\sum_{i=1}^{d} x_i^2}

    - **Angles** (:math:`\phi_1, \phi_2, \dots, \phi_{d-2}, \phi_{d-1}`):

      .. math::

          \begin{align}
            \phi_1 =& \text{atan2}(\sqrt{x_d^2 + x_{d-1}^2 + \dots + x_2^2, x_1}) \\
            \phi_2 =& \text{atan2}(\sqrt{x_d^2 + x_{d-1}^2 + \dots + x_3^2, x_2}) \\
            &\vdots \\
            \phi_{d-2} =& \text{atan2}(\sqrt{x_d^2 + x_{d-1}^2, x_{d-2}}) \\
            \phi_{d-1} =& \text{atan2}(x_d, x_{d-1})
          \end{align}

    Special handling is provided for angles to ensure they lie within the interval
    :math:`[0, 2\pi]`.

    References
    ----------

    .. [1] Wikipedia. "n-sphere" `Wiki Link <https://en.wikipedia.org/wiki/N-sphere>`_
    .. [2] Blumenson, L. E. (1960). "A Derivation of n-Dimensional Spherical Coordinates".
           The American Mathematical Monthly. 67 (1): 63–66.
           `JSTOR Link <https://doi.org/10.2307%2F2308932>`_


    Examples
    --------
    Convert a set of 3D Cartesian coordinates to spherical coordinates:

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.kernel.coordinate_system import cart2sph

        cart = np.array([[1, 1, 1], [0, 0, 1]])
        sph = cart2sph(cart)
        print(sph)
        # Output:
        # array([[1.73205081, 0.95531662, 0.78539816],
        #        [1.        , 0.        , 0.        ]])

    Convert a set of 2D Cartesian coordinates:

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.kernel.coordinate_system import cart2sph

        cart = np.array([[1, 1], [0, 1]])
        sph = cart2sph(cart)
        print(sph)
        # Output:
        # array([[1.41421356, 0.78539816],
        #        [1.        , 0.        ]])

    """
    n = cart.shape[0]
    d = cart.shape[1]
    sph = np.zeros((n, d))
    for i in range(n):
        sph[i, 0] = np.linalg.norm(cart[i, :])
        for j in range(d - 2):
            sph[i, j + 1] = np.arccos(cart[i, j] / np.linalg.norm(cart[i, j:]))
        if np.linalg.norm(cart[i, -2:]) == 0:
            sph[i, -1] = 0
        elif cart[i, -1] >= 0:
            sph[i, -1] = np.arccos(cart[i, -2] / np.linalg.norm(cart[i, -2:]))
        else:
            sph[i, -1] = 2 * np.pi - np.arccos(cart[i, -2] / np.linalg.norm(cart[i, -2:]))
    return sph


def sph2cart(sph):
    r"""
    Convert spherical coordinates to Cartesian coordinates.

    This function transforms a set of points in d-dimensional spherical coordinates to d-dimensional
    Cartesian coordinates. In Cartesian coordinates, each point is represented the d-dimensional
    signed-distance from given coordinate points to d mutually perpendicular, fixed hyperplanes [1]_.

    Parameters
    ----------
    sph : ndarray of shape (N, d)
        Spherical coordinates for N points in d-dimensional space. Each row represents the
        spherical coordinates of a point, where the first element is the radial distance, and the
        remaining elements are angles in radians.

    Returns
    -------
    cart : ndarray of shape (N, d)
        Cartesian coordinates for the N points. Each row represents the Cartesian coordinates of a
        point in d-dimensional space.


    Notes
    -----
    The conversion from spherical to Cartesian coordinates in n-dimensional space is performed using
    the following formulas [2]_:

    For a point with radial distance ( :math:`r` \ and angles ( :math:`\phi_1, \phi_2, \phi_3, \dots, \phi_{d-1}` ):

    - **Cartesian coordinates** :math:`\mathbf{x}_i \forall i = 1, \dots, d`:

      .. math::

          x_1 &= r \cos(\phi_1) \\
          x_2 &= r \sin(\phi_1) \cos(\phi_2) \\
          x_3 &= r \sin(\phi_1) \sin(\phi_2) \cos(\phi_3) \\
          &\vdots \\
          x_{d-1} &= r \left( \prod_{k=1}^{d-2} \sin(\phi_k) \right) \cos(\phi_{d-1}) \\
          x_d &= r \left( \prod_{k=1}^{d-1} \sin(\phi_k) \right)

    Special handling is provided to prevent numerical errors when angles are at critical values such as
    \( :math:`0, \pm\frac{\pi}{2}, \pi` \).

    References
    ----------

    .. [1] Wikipedia. "Hyperplane" `Wiki Link <https://en.wikipedia.org/wiki/Hyperplane>`_
    .. [2] Blumenson, L. E. (1960). "A Derivation of n-Dimensional Spherical Coordinates".
           The American Mathematical Monthly. 67 (1): 63–66.
           `JSTOR Link <https://doi.org/10.2307%2F2308932>`_

    Examples
    --------

    Convert a set of 2D spherical coordinates:

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.kernel.coordinate_system import sph2cart

        sph = np.array([[1.41421356, 0.78539816], [1.0, 0.0]])
        cart = sph2cart(sph)
        print(cart)
        # Output:
        # array([[1., 1.],
        #        [1., 0.]])

    Convert a set of 3D spherical coordinates to Cartesian coordinates:

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.kernel.coordinate_system import sph2cart

        sph = np.array([[1.73205081, 0.95531662, 0.78539816], [1.0, 0.0, 0.0]])
        cart = sph2cart(sph)
        print(cart)
        # Output:
        # array([[1., 1., 1.],
        #        [1., 0., 0.]])
    """
    n = sph.shape[0]
    d = sph.shape[1]
    cart = np.zeros((n, d))
    for i in range(n):
        if np.isclose(sph[i, 1], np.pi/2) or np.isclose(sph[i, 1], -np.pi/2) or np.isclose(sph[i, 1], (3/2)*np.pi):
            cart[i, 0] = 0
        else:
            cart[i, 0] = sph[i, 0] * np.cos(sph[i, 1])
        for j in range(max(d - 2, 0)):
            if np.any(np.isclose(sph[i, 1:j + 2], np.pi)) or np.any(np.isclose(sph[i, 1:j + 2], -np.pi)) or \
                    np.any(np.isclose(sph[i, 1:j + 2], 0)):
                cart[i, j + 1] = 0
            elif np.isclose(sph[i, j + 2], np.pi/2) or np.isclose(sph[i, j + 2], -np.pi/2) or \
                    np.isclose(sph[i, j + 2], (3/2)*np.pi):
                cart[i, j + 1] = 0
            else:
                p_ones = np.isclose(sph[i, 1:j+2], np.pi/2)
                n_ones = np.logical_or(np.isclose(sph[i, 1:j+2], -np.pi/2), np.isclose(sph[i, 1:j+2], (3/2)*np.pi))
                sins = np.sin(sph[i, 1:j+2])
                sins[p_ones] = 1
                sins[n_ones] = -1
                if np.isclose(sph[i, j + 2], 0):
                    cos = 1
                elif np.isclose(sph[i, j + 2], np.pi) or np.isclose(sph[i, j + 2], -np.pi):
                    cos = -1
                else:
                    cos = np.cos(sph[i, j + 2])
                cart[i, j + 1] = sph[i, 0] * np.prod(sins) * cos
        if np.any(np.isclose(sph[i, 1:], np.pi)) or np.any(np.isclose(sph[i, 1:], -np.pi)):
            cart[i, -1] = 0
        else:
            p_ones = np.isclose(sph[i, 1:], np.pi/2)
            n_ones = np.isclose(sph[i, 1:], -np.pi/2)
            sins = np.sin(sph[i, 1:])
            sins[p_ones] = 1
            sins[n_ones] = -1
            cart[i, -1] = sph[i, 0] * np.prod(sins)
    return cart



import numpy as np
import matplotlib.pyplot as plt
from abc import ABC, abstractmethod


class BaseCurve(ABC):
    """
    A base abstract class for curves in 2D or 3D space.

    Provides a consistent interface for evaluating, differentiating,
    computing curvature/torsion, arc length, etc. Also includes
    optional methods like dimension, bounding box, transformations,
    and plotting.
    """

    @abstractmethod
    def evaluate(self, t_values):
        """
        Evaluate the curve at given parametric values.
        Parameters
        ----------
        t_values : array-like
            Parametric values, typically in [0, 1].
        Returns
        -------
        np.ndarray
            Points on the curve at each parametric value.
        """
        pass

    @abstractmethod
    def derivative(self, t_values, order=1):
        """
        Compute the nth derivative of the curve.
        Parameters
        ----------
        t_values : array-like
            Parametric values in [0, 1].
        order : int, optional
            The order of the derivative (1 for first derivative, etc.).
        Returns
        -------
        np.ndarray
            Derivative values at each parametric value.
        """
        pass

    @abstractmethod
    def roc(self, t_values):
        """
        Compute the radius of curvature at specified parametric values.
        """
        pass

    @abstractmethod
    def torsion(self, t_values):
        """
        Compute the torsion at specified parametric values.
        """
        pass

    @abstractmethod
    def arc_length(self, t_start=0.0, t_end=1.0, num_points=100):
        """
        Approximate the arc length of the curve from t_start to t_end.
        """
        pass

    # -- Additional Optional Methods / Properties --

    @property
    @abstractmethod
    def dimension(self):
        """
        Return the dimension of the curve (2 for 2D, 3 for 3D).
        """
        pass

    def bounding_box(self, num_samples=100):
        """
        Return an axis-aligned bounding box of the curve over its parameter range.
        This implementation is a naive sample-based approach; override for exact if needed.
        """
        # By default, assume domain is [0,1], override if your curve has different domain
        t_values = np.linspace(0, 1, num_samples)
        points = self.evaluate(t_values)
        min_coords = np.min(points, axis=0)
        max_coords = np.max(points, axis=0)
        return min_coords, max_coords

    def is_closed(self, tol=1e-7):
        """
        Check if the curve is closed within some tolerance (start point == end point).
        """
        start_point = self.evaluate([0])[0]
        end_point = self.evaluate([1])[0]
        return np.allclose(start_point, end_point, atol=tol)

    def transform(self, matrix):
        """
        Apply a transformation matrix to the curve's underlying data.
        By default, raise NotImplementedError; each subclass must define its own transformation logic.
        """
        raise NotImplementedError(
            "Subclass must override transform method to handle transformations."
        )

    def plot(self, num_points=100, t_start=0.0, t_end=1.0, ax=None, **plot_kwargs):
        """
        Quick 2D plot of the curve using matplotlib.

        Parameters
        ----------
        num_points : int
            Number of sample points.
        t_start : float
            Start parameter (default 0.0).
        t_end : float
            End parameter (default 1.0).
        ax : matplotlib.axes.Axes
            If provided, draw on the given axes. Otherwise, create a new figure.
        plot_kwargs : dict
            Additional keyword arguments passed to `matplotlib.pyplot.plot`.
        """
        if self.dimension < 2:
            raise ValueError("Cannot plot a curve with dimension < 2.")

        # Evaluate points
        t_values = np.linspace(t_start, t_end, num_points)
        points = self.evaluate(t_values)  # shape: (num_points, dimension)

        if ax is None:
            fig, ax = plt.subplots()

        # If dimension=2, we can do a direct 2D plot
        if self.dimension == 2:
            ax.plot(points[:, 0], points[:, 1], **plot_kwargs)
            ax.set_aspect('equal', adjustable='datalim')
            ax.set_title("2D Curve")
        elif self.dimension == 3:
            # For a 3D curve, you'd need a 3D axis
            from mpl_toolkits.mplot3d import Axes3D  # or use matplotlib >= 3.2 with fig.add_subplot(projection='3d')
            if not hasattr(ax, 'plot3D'):
                # auto-create a 3D axis if user didn’t provide one
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
            ax.plot3D(points[:, 0], points[:, 1], points[:, 2], **plot_kwargs)
            ax.set_title("3D Curve")
        else:
            raise NotImplementedError("Plotting only implemented for 2D or 3D curves.")

        return ax



# test_nurbs_curve.py

import pytest
import numpy as np

from svv.forest.connect.base import Curve # the base curve class
from svv.forest.connect.nurbs import NURBSCurve # The nurbs curve class


def test_constructor_valid_linear_2d():
    """
    Test constructing a 2D linear NURBS curve with 2 control points (straight line).
    For degree=1, we need knot_vector of length = N + degree + 1 = 2 + 1 + 1 = 4.
    """
    control_points = [[0, 0], [2, 2]]
    weights = [1, 1]  # each control point has weight=1
    knot_vector = [0, 0, 1, 1]  # typical 'clamped' form for linear
    degree = 1

    curve = NURBSCurve(control_points, weights, knot_vector, degree)
    assert isinstance(curve, NURBSCurve)
    assert curve.dimension == 2
    assert len(curve.control_points) == 2
    assert len(curve.weights) == 2
    assert len(curve.knot_vector) == 4


def test_constructor_invalid_weight_length():
    """
    control_points and weights must have same length.
    """
    control_points = [[0, 0], [1, 1]]
    weights = [1]  # mismatched length
    knot_vector = [0, 0, 1, 1]
    degree = 1

    with pytest.raises(ValueError, match="must have the same length"):
        NURBSCurve(control_points, weights, knot_vector, degree)


def test_constructor_invalid_knot_length():
    """
    For N=3 control points, degree=2 => required knot length = N+degree+1 = 6,
    so giving anything else should raise an error.
    """
    control_points = [[0, 0], [1, 1], [2, 2]]
    weights = [1, 1, 1]
    # This is the wrong length (should be 3+2+1=6). Let's use 5 instead:
    knot_vector = [0, 0, 0, 1, 1]
    degree = 2

    with pytest.raises(ValueError, match="knot_vector length must be"):
        NURBSCurve(control_points, weights, knot_vector, degree)


def test_constructor_non_decreasing_knot():
    """
    The knot vector must be non-decreasing.
    """
    control_points = [[0, 0], [1, 1], [2, 2]]
    weights = [1, 1, 1]
    # This has a decreasing portion: [0, 0, 1, 0, 1, 2]
    knot_vector = [0, 0, 1, 0, 1, 2]
    degree = 2

    with pytest.raises(ValueError, match="must be non-decreasing"):
        NURBSCurve(control_points, weights, knot_vector, degree)


def test_linear_nurbs_evaluate():
    """
    For a linear NURBS curve (degree=1) with 2 points, weights=1 => it's a straight line.
    We check that evaluate(t) is a linear interpolation between (0,0) and (2,2).
    """
    control_points = [[0, 0], [2, 2]]
    weights = [1, 1]
    knot_vector = [0, 0, 1, 1]
    degree = 1

    curve = NURBSCurve(control_points, weights, knot_vector, degree)

    # Evaluate at t=0 => (0,0), t=1 => (2,2), t=0.5 => (1,1)
    t_values = [0.0, 0.5, 1.0]
    pts = curve.evaluate(t_values)
    expected = np.array([[0, 0],
                         [1, 1],
                         [2, 2]])
    np.testing.assert_allclose(pts, expected, atol=1e-8)


def test_evaluate_scalar_t():
    """
    Check that passing a single scalar t also works and returns shape (1, dimension).
    """
    control_points = [[0, 0], [1, 0]]
    weights = [1, 1]
    knot_vector = [0, 0, 1, 1]
    degree = 1

    curve = NURBSCurve(control_points, weights, knot_vector, degree)
    point = curve.evaluate(0.5)
    # Should return shape (1,2), midpoint between (0,0) and (1,0) => (0.5, 0).
    assert point.shape == (1, 2)
    np.testing.assert_allclose(point, [[0.5, 0.0]], atol=1e-8)


def test_dimension_property():
    """
    Confirm dimension matches the second axis of control_points.
    """
    control_points_2d = [[0, 0], [1, 1], [2, 2]]
    weights_2d = [1, 2, 1]
    knot_vector_2d = [0, 0, 0, 1, 1, 1]
    curve_2d = NURBSCurve(control_points_2d, weights_2d, knot_vector_2d, degree=2)
    assert curve_2d.dimension == 2

    control_points_3d = [[0, 0, 0], [1, 1, 1], [2, 0, 2]]
    weights_3d = [1, 1, 1]
    knot_vector_3d = [0, 0, 0, 1, 1, 1]
    curve_3d = NURBSCurve(control_points_3d, weights_3d, knot_vector_3d, degree=2)
    assert curve_3d.dimension == 3


def test_is_subclass_of_curve():
    """
    Check that NURBSCurve is indeed a subclass of Curve.
    """
    control_points = [[0, 0], [2, 2]]
    weights = [1, 1]
    knot_vector = [0, 0, 1, 1]
    degree = 1
    curve = NURBSCurve(control_points, weights, knot_vector, degree)
    assert isinstance(curve, Curve)


def test_not_implemented_methods():
    """
    derivative, roc, torsion, arc_length, transform should raise NotImplementedError
    in the current version of NURBSCurve.
    """
    control_points = [[0, 0], [2, 2]]
    weights = [1, 1]
    knot_vector = [0, 0, 1, 1]
    degree = 1
    curve = NURBSCurve(control_points, weights, knot_vector, degree)

    t_vals = [0.0, 0.5, 1.0]
    with pytest.raises(NotImplementedError):
        curve.derivative(t_vals)
    with pytest.raises(NotImplementedError):
        curve.roc(t_vals)
    with pytest.raises(NotImplementedError):
        curve.torsion(t_vals)
    with pytest.raises(NotImplementedError):
        curve.arc_length(0.0, 1.0)
    with pytest.raises(NotImplementedError):
        curve.transform(np.eye(3))


def test_bounding_box_linear():
    """
    bounding_box calls evaluate internally.
    For the line from (0,0) to (2,2), the bounding box should be:
      min => (0,0), max => (2,2).
    """
    control_points = [[0, 0], [2, 2]]
    weights = [1, 1]
    knot_vector = [0, 0, 1, 1]
    degree = 1
    curve = NURBSCurve(control_points, weights, knot_vector, degree)

    min_pt, max_pt = curve.bounding_box(num_samples=10)
    np.testing.assert_allclose(min_pt, [0, 0])
    np.testing.assert_allclose(max_pt, [2, 2])


def test_is_closed():
    """
    By default, is_closed() checks start_point vs end_point at t=0 vs t=1.
    A line from (0,0) to (2,2) is not closed => False.
    """
    control_points = [[0, 0], [2, 2]]
    weights = [1, 1]
    knot_vector = [0, 0, 1, 1]
    degree = 1
    curve = NURBSCurve(control_points, weights, knot_vector, degree)
    assert curve.is_closed() is False

    # If we forcibly made them the same, is_closed() would be True
    control_points2 = [[0, 0], [0, 0]]  # start == end
    weights2 = [1, 1]
    knot_vector2 = [0, 0, 1, 1]
    curve2 = NURBSCurve(control_points2, weights2, knot_vector2, degree=1)
    assert curve2.is_closed() is True



import numpy
from copy import deepcopy
from svv.tree.tree import Tree
from svv.tree.data.data import TreeData
from svv.tree.collision.tree_collision import tree_collision
from svv.forest.connect.geodesic import geodesic_constructor
from svv.forest.connect.forest_connection import ForestConnection
from svv.visualize.forest.show import show


class Forest(object):
    def __init__(self, **kwargs):
        """
        The Forest class defines collections of synthetic vascular tree
        structures that are used to abstract the physical representations
        of generated interpenetrating vascular networks.

        Parameters
        ----------
        kwargs : dict
            A dictionary of keyword arguments to be passed.
            Keyword arguments include:
                domain : svtoolkit.domain.Domain
                    A domain object that defines the spatial region in which vascular
                    trees are generated.
                n_networks : int
                    The number of vascular networks to be generated.
                n_trees_per_network : list of int
                    The number of trees to be generated per network.
                start_points : list of numpy.ndarray
                    A list of starting points for the trees.
                directions : list of numpy.ndarray
                    A list of directions for the trees.
                physical_clearance : float
                    The physical clearance of any vessel from the domain and other vessels.
                compete : bool
                    A flag that indicates whether trees should compete for space.
        """
        self.networks = []
        self.connections = None
        self.domain = kwargs.get('domain', None)
        self.n_networks = kwargs.get('n_networks', 1)
        self.n_trees_per_network = kwargs.get('n_trees_per_network', [2])
        self.start_points = kwargs.get('start_points', None)
        self.directions = kwargs.get('directions', None)
        self.physical_clearance = kwargs.get('physical_clearance', 0.0)
        self.compete = kwargs.get('compete', False)
        self.geodesic = None
        self.convex = None
        if isinstance(self.start_points, type(None)):
            self.start_points = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
        if isinstance(self.directions, type(None)):
            self.directions = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
        networks = []
        for i in range(self.n_networks):
            network = []
            for j in range(self.n_trees_per_network[i]):
                tree = Tree()
                tree.physical_clearance = self.physical_clearance
                network.append(tree)
            networks.append(network)
        self.networks = networks

    def set_domain(self, domain, convexity_tolerance=1e-2):
        """
        Set the domain for the forest. The domain should be an implicit
        function of the form f(X) = scalar, where X is a cartesian
        coordinate in space.

        Parameters
        ----------
        domain : svtoolkit.domain.Domain
            A domain object that defines the spatial region in which vascular
            trees are generated.
        convexity_tolerance : float
            The tolerance for the convexity of the domain.
        """
        self.domain = domain
        self.geodesic = geodesic_constructor(domain)
        self.convex = numpy.isclose(domain.convexity, 1.0, atol=convexity_tolerance)
        for network in self.networks:
            for tree in network:
                tree.set_domain(domain, convexity_tolerance)
        return None

    def set_roots(self, *args, **kwargs):
        """
        Set the root point of the forest.

        Parameters
        ----------
        start_points : list of numpy.ndarray
            A list of starting points for the trees.
        directions : list of numpy.ndarray
            A list of directions for the trees.
        kwargs : dict
        """
        if len(args) == 0:
            kwargs['start_points'] = self.start_points
            kwargs['directions'] = self.directions
        elif len(args) == 1:
            kwargs['start_points'] = args[0]
            kwargs['directions'] = self.directions
        elif len(args) == 2:
            kwargs['start_points'] = args[0]
            kwargs['directions'] = args[1]
        else:
            raise ValueError("Too many arguments.")
        tmp_roots = []
        tmp_root_maps = []
        tmp_connectivities = []
        tmp_kdtms = []
        tmp_hnsw_trees = []
        tmp_hnsw_tree_ids = []
        tmp_probabilities = []
        tmp_tree_scales = []
        for i in range(self.n_networks):
            for j in range(self.n_trees_per_network[i]):
                success = False
                while not success:
                    tmp_root, tmp_root_map,tmp_connecivity,tmp_kdtm,tmp_hnsw_tree,tmp_hnsw_tree_id,tmp_probability,tmp_tree_scale = self.networks[i][j].set_root(start=kwargs['start_points'][i][j],
                                                                          direction=kwargs['directions'][i][j],
                                                                          inplace=False)
                    collisions = []
                    for k in range(len(tmp_roots)):
                        root_collision = tree_collision(tmp_roots[k], tmp_root, clearance=self.physical_clearance)
                        collisions.append(root_collision)
                        if root_collision:
                            break
                    if not any(collisions):
                        success = True
                tmp_roots.append(tmp_root)
                tmp_root_maps.append(tmp_root_map)
                tmp_connectivities.append(tmp_connecivity)
                tmp_kdtms.append(tmp_kdtm)
                tmp_hnsw_trees.append(tmp_hnsw_tree)
                tmp_hnsw_tree_ids.append(tmp_hnsw_tree_id)
                tmp_probabilities.append(tmp_probability)
                tmp_tree_scales.append(tmp_tree_scale)
        for i in range(self.n_networks):
            for j in range(self.n_trees_per_network[i]):
                self.networks[i][j].preallocate[0, :] = tmp_roots.pop(0)
                self.networks[i][j].data = self.networks[i][j].preallocate[:1, :]
                self.networks[i][j].vessel_map.update(tmp_root_maps.pop(0))
                self.networks[i][j].n_terminals = 1
                self.networks[i][j].max_distal_node = 1
                self.networks[i][j].connectivity = tmp_connectivities.pop(0)
                #self.networks[i][j].kdtm = tmp_kdtms.pop(0)
                self.networks[i][j].hnsw_tree = tmp_hnsw_trees.pop(0)
                self.networks[i][j].hnsw_tree_id = tmp_hnsw_tree_ids.pop(0)
                self.networks[i][j].probability = tmp_probabilities.pop(0)
                self.networks[i][j].tree_scale = tmp_tree_scales.pop(0)
                self.networks[i][j].midpoints = (self.networks[i][j].data[:, 0:3] + self.networks[i][j].data[:, 3:6])/2
                self.networks[i][j].segment_count = 1

    def add(self, *args, **kwargs):
        """
        Add a tree to the forest.
        """
        decay_probability = kwargs.pop('decay_probability', 0.9)
        self.connections = None
        if len(args) == 0:
            n_vessels = 1
        else:
            n_vessels = args[0]
        progress = kwargs.get('progress', False)
        network_id = kwargs.get('network_id', None)
        if isinstance(network_id, type(None)):
            network_id = list(range(self.n_networks))
        for k in range(n_vessels):
            new_vessels = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
            added_vessels = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
            new_vessel_maps = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
            new_connectivity = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
            new_inds = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
            new_mesh_cell = [[None for j in range(self.n_trees_per_network[i])] for i in range(self.n_networks)]
            for i in network_id:
                for j in range(self.n_trees_per_network[i]):
                    success = False
                    while not success:
                        #tmp_new_vessels, tmp_added_vessels, tmp_new_vessel_map, _, _, tmp_nonconvex_count = self.networks[i][j].add(inplace=False)
                        results = self.networks[i][j].add(inplace=False)
                        change_i, change_j, new_tmp_data, old_tmp_data, tmp_new_vessel_map, connectivity, inds, mesh_cell, tmp_added_vessels = results
                        #tmp_new_vessels = numpy.vstack([self.networks[i][j].data, tmp_added_vessels[0], tmp_added_vessels[1]])
                        change_i = numpy.array(change_i, dtype=int)
                        change_j = numpy.array(change_j, dtype=int)
                        self.networks[i][j].preallocate[self.networks[i][j].segment_count, :] = tmp_added_vessels[0]
                        self.networks[i][j].preallocate[self.networks[i][j].segment_count+1, :] = tmp_added_vessels[1]
                        self.networks[i][j].preallocate[change_i, change_j] = numpy.array(new_tmp_data)
                        self.networks[i][j].data = self.networks[i][j].preallocate[:self.networks[i][j].segment_count+2, :]
                        self.networks[i][j].preallocate_midpoints[:self.networks[i][j].segment_count+2, :] = (self.networks[i][j].data[:, 0:3] +
                                                                                                              self.networks[i][j].data[:, 3:6])/2
                        self.networks[i][j].midpoints = self.networks[i][j].preallocate_midpoints[:self.networks[i][j].segment_count + 2, :]
                        #tmp_new_vessels[change_i, change_j] = numpy.array(new_tmp_data)
                        tmp_new_vessels = self.networks[i][j].data
                        collisions = []
                        for h in range(self.n_networks):
                            for w in range(self.n_trees_per_network[h]):
                                if h == i and w == j:
                                    continue
                                for q in range(len(tmp_added_vessels)):
                                    if isinstance(new_vessels[h][w], type(None)):
                                        collision = tree_collision(self.networks[h][w].data, tmp_added_vessels[q],
                                                                   clearance=self.physical_clearance)
                                    else:
                                        collision = tree_collision(new_vessels[h][w], tmp_added_vessels[q],
                                                                   clearance=self.physical_clearance)
                                    collisions.append(collision)
                                    if collision:
                                        print("Collision detected {}-{}.".format(h, q))
                                        break
                                if any(collisions):
                                    break
                            if any(collisions):
                                break
                        if not any(collisions):
                            success = True
                            new_vessels[i][j] = tmp_new_vessels
                            added_vessels[i][j] = tmp_added_vessels
                            new_vessel_maps[i][j] = tmp_new_vessel_map
                            new_connectivity[i][j] = connectivity
                            new_inds[i][j] = inds
                            new_mesh_cell[i][j] = mesh_cell
                        else:
                            success = False
                            self.networks[i][j].preallocate[self.networks[i][j].segment_count, :] = numpy.nan
                            self.networks[i][j].preallocate[self.networks[i][j].segment_count + 1, :] = numpy.nan
                            self.networks[i][j].preallocate[change_i, change_j] = numpy.array(old_tmp_data)
                            self.networks[i][j].data = self.networks[i][j].preallocate[:self.networks[i][j].segment_count + 2,:]
                            self.networks[i][j].preallocate_midpoints[:self.networks[i][j].segment_count + 2, :] = (self.networks[i][j].data[:,0:3] +
                                                                                                                    self.networks[i][j].data[:,3:6]) / 2
                            self.networks[i][j].midpoints = self.networks[i][j].preallocate_midpoints[
                                                            :self.networks[i][j].segment_count, :]
            for i in network_id:
                for j in range(self.n_trees_per_network[i]):
                    #_data = TreeData(new_vessels[i][j].shape)
                    #_data[:, :] = new_vessels[i][j][:, :]
                    #self.networks[i][j].data = TreeData.from_array(new_vessels[i][j])
                    #self.networks[i][j].vessel_map.update(new_vessel_maps[i][j])
                    for key in new_vessel_maps[i][j].keys():
                        if key in self.networks[i][j].vessel_map.keys():
                            self.networks[i][j].vessel_map[key]['upstream'].extend(new_vessel_maps[i][j][key]['upstream'])
                            self.networks[i][j].vessel_map[key]['downstream'].extend(new_vessel_maps[i][j][key]['downstream'])
                        else:
                            self.networks[i][j].vessel_map[key] = deepcopy(new_vessel_maps[i][j][key])
                    self.networks[i][j].connectivity = new_connectivity[i][j]
                    self.networks[i][j].max_distal_node += 2
                    if new_mesh_cell[i][j] >= 0:
                        self.networks[i][j].probability[mesh_cell] *= decay_probability
                        self.networks[i][j].probability = self.networks[i][j].probability / self.networks[i][j].probability.sum()
                        self.networks[i][j].domain.cumulative_probability = numpy.cumsum(self.networks[i][j].probability)
                    self.networks[i][j].hnsw_tree.replace(
                        ((new_vessels[i][j][new_inds[i][j][0], 0:3] + new_vessels[i][j][new_inds[i][j][0], 3:6]) / 2).reshape(1, 3).astype(
                            numpy.float32), numpy.array([new_inds[i][j][0]]))
                    self.networks[i][j].hnsw_tree.add_items(
                        ((new_vessels[i][j][new_inds[i][j][1], 0:3] + new_vessels[i][j][new_inds[i][j][1], 3:6]) / 2).reshape(1, 3).astype(
                            numpy.float32), numpy.array([new_inds[i][j][1]]))
                    self.networks[i][j].hnsw_tree.add_items(
                        ((new_vessels[i][j][new_inds[i][j][2], 0:3] + new_vessels[i][j][new_inds[i][j][2], 3:6]) / 2).reshape(1, 3).astype(
                            numpy.float32), numpy.array([new_inds[i][j][2]]))
                    self.networks[i][j].tree_scale = self.networks[i][j].new_tree_scale
                    self.networks[i][j].n_terminals += 1
                    self.networks[i][j].segment_count += 2

    def connect(self, *args, **kwargs):
        self.connections = ForestConnection(self)
        if len(args) == 0:
            self.connections.solve(1, **kwargs)
        elif len(args) == 1 and isinstance(args[0], int):
            self.connections.solve(*args, **kwargs)
        else:
            raise ValueError("Invalid arguments. Optional integer argument for default Bezier polynomial degree.")

    def show(self, **kwargs):
        return show(self, **kwargs)

    def export_solid(self, outdir=None, shell_thickness=0.0):
        """
        Export the domain as a solid object.
        """
        if isinstance(outdir, type(None)):
            outdir = '3d_tmp'
        for i in range(self.n_networks):
            for j in range(self.n_trees_per_network[i]):
                self.networks[i][j].export_solid(outdir=outdir, shell_thickness=shell_thickness)


from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import numpy
import sys
import os
import platform
import subprocess
from urllib.request import urlretrieve
import shutil
import tarfile
import json
import stat
import multiprocessing

num_cores = multiprocessing.cpu_count() // 2

from setuptools.command.build_ext import build_ext
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

class BDistWheelCmd(_bdist_wheel):
    def run(self):
        # Ensure build_ext (which calls build_mmg) runs first
        #build_mmg()
        self.run_command("build_ext")
        super().run()
        
def get_filename_without_ext(abs_path):
    base_name = os.path.basename(abs_path)      # e.g. "file.txt"
    file_name_no_ext = os.path.splitext(base_name)[0]  # e.g. "file"
    return file_name_no_ext

def remove_directory_tree(directory_path):
    """
    Removes the directory at `directory_path` along with all its files/subdirectories.
    """
    if os.path.exists(directory_path) and os.path.isdir(directory_path):
        shutil.rmtree(directory_path)
    else:
        print(f"Directory does not exist or is not a directory: {directory_path}")

def find_executables(top_level_folder):
    """
    Return a list of absolute paths for all executables found
    within 'top_level_folder' (recursively).

    On Windows:
        - Treat files with certain extensions (.exe, .bat, .cmd, .ps1)
          as executables.

    On Linux/macOS:
        - Check if the file has the executable bit set.

    Parameters:
        top_level_folder (str): Path to the directory where we search for executables.

    Returns:
        list of str: A list of absolute file paths that are considered executable.
    """
    # Normalize the top folder path
    top_level_folder = os.path.abspath(top_level_folder)

    # Define criteria:
    windows_exts = {".exe", ".bat", ".cmd", ".ps1"}
    is_windows = platform.system().lower().startswith("win")

    executables = []

    for root, dirs, files in os.walk(top_level_folder):
        for filename in files:
            full_path = os.path.join(root, filename)
            if is_windows:
                # On Windows, check extension
                _, ext = os.path.splitext(filename)
                if ext.lower() in windows_exts:
                    executables.append(os.path.abspath(full_path))
            else:
                # On Linux/macOS, check executable bit
                # (Also ensure it's not a directory, just in case)
                mode = os.stat(full_path).st_mode
                if (mode & stat.S_IXUSR) and not os.path.isdir(full_path):
                    executables.append(os.path.abspath(full_path))

    return executables

def find_vs_installations():
    # Modify this path if vswhere.exe is in a different place
    vswhere_path = r"C:\Program Files (x86)\Microsoft Visual Studio\Installer\vswhere.exe"

    if not os.path.exists(vswhere_path):
        print("vswhere.exe not found at:", vswhere_path)
        return []

    # Call vswhere to get info about all VS installations in JSON format
    vswhere_cmd = [
        vswhere_path,
        "-all",  # list all VS instances
        "-requires", "Microsoft.Component.MSBuild",  # only show VS installs with MSBuild
        "-format", "json"  # output in JSON
    ]

    result = subprocess.run(vswhere_cmd, capture_output=True, text=True)
    if result.returncode != 0:
        print("vswhere failed or returned an error.")
        return []

    output = result.stdout.strip()
    if not output:
        print("No Visual Studio installations found by vswhere.")
        return []

    try:
        data = json.loads(output)
    except json.JSONDecodeError:
        print("Failed to parse vswhere JSON output.")
        return []

    return data


def pick_visual_studio_generator():
    """
    Looks up installed Visual Studio versions via vswhere and picks
    the highest known generator that CMake can use, e.g.
    - "Visual Studio 17 2022" (for VS 2022)
    - "Visual Studio 16 2019" (for VS 2019)

    Returns the generator string or None if none found.
    """
    # This dictionary maps partial version info to the CMake generator name
    # Adjust as needed if you want more logic:
    vs_generators = {
        "17.": "Visual Studio 17 2022",
        "16.": "Visual Studio 16 2019",
        "15.": "Visual Studio 15 2017",
    }

    installations = find_vs_installations()
    if not installations:
        return None

    # We'll keep track of the best (highest) version found
    best_gen = None
    best_version_num = 0.0

    for inst in installations:
        # vs["catalog"]["productDisplayVersion"] might look like "17.5.33424.131"
        version_str = inst.get("catalog", {}).get("productDisplayVersion", "")
        # Attempt to parse out the major version (17, 16, 15, etc.)
        if version_str:
            # e.g., "17.4.1" => major = 17
            try:
                major_str = version_str.split(".")[0]  # '17'
                major = int(major_str)
            except ValueError:
                major = 0  # fallback

            # If we have a known generator string for this major version, use it
            # Or you might parse vs_generators keys more systematically
            if major >= 17 and "17." in version_str:
                if major > best_version_num:
                    best_gen = vs_generators["17."]
                    best_version_num = major
            elif major == 16 and "16." in version_str:
                if major > best_version_num:
                    best_gen = vs_generators["16."]
                    best_version_num = major
            elif major == 15 and "15." in version_str:
                if major > best_version_num:
                    best_gen = vs_generators["15."]
                    best_version_num = major

    return best_gen


def build_mmg(num_cores=None):
    if num_cores is None:
        num_cores = os.cpu_count() or 1

    # Make sure cmake is on PATH
    if shutil.which("cmake") is None:
        raise RuntimeError("CMake is not installed or not on the PATH.")

    download_url_mmg = "https://github.com/MmgTools/mmg/archive/refs/tags/v5.8.0.tar.gz"
    tarball_path_mmg = "mmg.tar.gz"
    source_extract_root = "mmg"

    # Download mmg if not present
    if not os.path.exists(tarball_path_mmg):
        print(f"Downloading {download_url_mmg}...")
        urlretrieve(download_url_mmg, tarball_path_mmg)

    # Extract mmg if not already extracted
    if not os.path.exists(source_extract_root):
        print("Extracting mmg...")
        with tarfile.open(tarball_path_mmg, "r:gz") as t:
            t.extractall(source_extract_root)

    # Typically the archive extracts into a folder named "mmg-5.8.0" under "mmg/"
    subdirs = os.listdir(source_extract_root)
    if not subdirs:
        raise RuntimeError("No files found after extracting mmg archive.")
    mmg_subdir = os.path.join(source_extract_root, subdirs[0])

    # Prepare build directory
    build_dir_mmg = os.path.abspath(os.path.join("bin", "mmg"))
    os.makedirs(build_dir_mmg, exist_ok=True)

    # Build up our cmake configure command
    cmake_cmd = ["cmake"]

    # On Windows, pick a Visual Studio generator if possible
    if platform.system().lower().startswith("win"):
        vs_generator = pick_visual_studio_generator()
        if vs_generator:
            cmake_cmd += ["-G", vs_generator]
        else:
            print("No suitable Visual Studio found, falling back to default generator or NMake.")
            # cmake_cmd += ["-G", "NMake Makefiles"]  # optional fallback

    # Add standard arguments
    cmake_cmd += [
        "-DCMAKE_BUILD_TYPE=Release",
        "-B", build_dir_mmg,
        "-S", mmg_subdir
    ]

    # Run configure step
    print("Configuring mmg with CMake:", " ".join(cmake_cmd))
    try:
        subprocess.check_call(cmake_cmd)
    except subprocess.CalledProcessError as e:
        raise RuntimeError("CMake configure failed for mmg.") from e

    # Run build step
    build_cmd = [
        "cmake",
        "--build", build_dir_mmg,
        "--parallel", str(num_cores)
    ]

    # For multi-config generators (Visual Studio, Xcode), we must specify --config Release
    # to ensure a Release build.
    if platform.system().lower().startswith("win"):
        # or detect if vs_generator is set if you want to be more precise
        build_cmd += ["--config", "Release"]

    print("Building mmg with CMake:", " ".join(build_cmd))

    try:
        subprocess.check_call(build_cmd)
    except subprocess.CalledProcessError as e:
        raise RuntimeError("CMake build failed for mmg.") from e

    print("MMG build completed successfully!")

    install_tmp_prefix = os.path.join("svv", "tmp")
    os.makedirs(install_tmp_prefix, exist_ok=True)
    install_cmd = [
        "cmake",
        "--install", build_dir_mmg,
        "--prefix", os.path.abspath(install_tmp_prefix),
    ]

    # For multi-configuration generators on Windows (like Visual Studio),
    # specify `--config Release` explicitly if you built in Release mode.
    if platform.system().lower().startswith("win"):
        install_cmd += ["--config", "Release"]

    print("Installing mmg with CMake:", " ".join(install_cmd))
    subprocess.check_call(install_cmd)
    print(f"mmg executables have been installed into: {install_tmp_prefix}")

    print("Copying executables and cleaning up")
    install_prefix = os.path.join("svv", "bin")
    os.makedirs(install_prefix, exist_ok=True)
    init_file = os.path.join('svv', 'bin', '__init__.py')
    if not os.path.isfile(init_file):
        with open(init_file, 'w'):
            pass
    executables = find_executables(install_tmp_prefix)
    basenames = ["mmg2d_O3", "mmg3d_O3", "mmgs_O3"]
    executables = [exe for exe in executables if get_filename_without_ext(exe) in basenames]
    for exe in executables:
        shutil.copy2(exe, install_prefix)
    remove_directory_tree(install_tmp_prefix)

    print('Remove Source, Archive, and Build directories')
    if os.path.isfile(tarball_path_mmg):
        os.remove(tarball_path_mmg)
    remove_directory_tree(build_dir_mmg)
    remove_directory_tree(source_extract_root)


def build_0d(num_cores=None):
    if num_cores is None:
        num_cores = os.cpu_count() or 1

    # Make sure cmake is on PATH
    if shutil.which("cmake") is None:
        raise RuntimeError("CMake is not installed or not on the PATH.")

    download_url_0d = "https://github.com/SimVascular/svZeroDSolver/archive/refs/tags/v2.0.tar.gz"
    tarball_path_0d = "svZeroDSolver.tar.gz"
    source_path_0d = os.path.abspath("svZeroDSolver")

    # Build up our cmake configure command
    cmake_cmd = ["cmake"]

    try:
        if not os.path.exists(tarball_path_0d):
            print(f"Downloading {download_url_0d}...")
            urlretrieve(download_url_0d, tarball_path_0d)
    except Exception as e:
        raise RuntimeError("Error downloading solver archives.") from e

    try:
        if not os.path.exists(source_path_0d):
            with tarfile.open(tarball_path_0d, "r:gz") as t:
                t.extractall(source_path_0d)
    except Exception as e:
        raise RuntimeError("Error extracting solver archives.") from e

    build_dir_0d = os.path.abspath("tmp/solver-0d")

    # Add standard arguments
    cmake_cmd += [
        "-DCMAKE_BUILD_TYPE=Release",
        "-B", build_dir_0d,
        "-S", mmg_subdir
    ]

    # Run build step
    build_cmd = [
        "cmake",
        "--build", build_dir_0d,
        "--parallel", str(num_cores)
    ]
class DownloadAndBuildExt(build_ext):
    def run(self):
        #------------------------------
        # Get MMG Remeshing Source Code
        #------------------------------
        build_mmg()
        # -----------------------------
        # 1. Download: handle errors
        # -----------------------------
        if not platform.system().lower().startswith("win"):
            """
            download_url_0d = "https://github.com/SimVascular/svZeroDSolver/archive/refs/tags/v2.0.tar.gz"
            download_url_1d = "https://github.com/SimVascular/svOneDSolver/archive/refs/tags/c9caded.tar.gz"
            download_url_3d = "https://github.com/SimVascular/svMultiPhysics/archive/refs/tags/March_2025.tar.gz"

            tarball_path_0d = "svZeroDSolver.tar.gz"
            tarball_path_1d = "svOneDSolver.tar.gz"
            tarball_path_3d = "svMultiPhysics.tar.gz"

            try:
                if not os.path.exists(tarball_path_0d):
                    print(f"Downloading {download_url_0d}...")
                    urlretrieve(download_url_0d, tarball_path_0d)
                if not os.path.exists(tarball_path_1d):
                    print(f"Downloading {download_url_1d}...")
                    urlretrieve(download_url_1d, tarball_path_1d)
                if not os.path.exists(tarball_path_3d):
                    print(f"Downloading {download_url_3d}...")
                    urlretrieve(download_url_3d, tarball_path_3d)
            except Exception as e:
                raise RuntimeError("Error downloading solver archives.") from e

            # -----------------------------
            # 2. Extract: handle errors
            # -----------------------------
            source_path_0d = os.path.abspath("svZeroDSolver")
            source_path_1d = os.path.abspath("svOneDSolver")
            source_path_3d = os.path.abspath("svMultiPhysics")

            try:
                if not os.path.exists(source_path_0d):
                    with tarfile.open(tarball_path_0d, "r:gz") as t:
                        t.extractall(source_path_0d)
                if not os.path.exists(source_path_1d):
                    with tarfile.open(tarball_path_1d, "r:gz") as t:
                        t.extractall(source_path_1d)
                if not os.path.exists(source_path_3d):
                    with tarfile.open(tarball_path_3d, "r:gz") as z:
                        z.extractall(source_path_3d)
            except Exception as e:
                raise RuntimeError("Error extracting solver archives.") from e

            # -----------------------------
            # 3. Configure & build with CMake
            # -----------------------------
            build_dir    = os.path.abspath("bin")
            build_dir_0d = os.path.abspath("bin/solver-0d")
            build_dir_1d = os.path.abspath("bin/solver-1d")
            build_dir_3d = os.path.abspath("bin/solver-3d")

            os.makedirs(build_dir_0d, exist_ok=True)
            os.makedirs(build_dir_1d, exist_ok=True)
            os.makedirs(build_dir_3d, exist_ok=True)

            # For each solver, configure and build. Catch errors if subprocess fails.
            try:
                subprocess.check_call(["cmake", f"-B{build_dir_0d}", f"-S{os.path.join(source_path_0d, 'svZeroDSolver-2.0')}"])
                subprocess.check_call(["cmake", f"-B{build_dir_1d}", f"-S{os.path.join(source_path_1d, 'svOneDSolver-c9caded')}"])
                subprocess.check_call(["cmake", f"-B{build_dir_3d}", f"-S{os.path.join(source_path_3d, 'svMultiPhysics-March_2025')}"])

                subprocess.check_call(["cmake", "--build", build_dir_0d, "--parallel", str(num_cores)])
                subprocess.check_call(["cmake", "--build", build_dir_1d, "--parallel", str(num_cores)])
                subprocess.check_call(["cmake", "--build", build_dir_3d, "--parallel", str(num_cores)])
            except subprocess.CalledProcessError as e:
                raise RuntimeError("CMake configure or build failed for one of the solvers.") from e

            # -----------------------------
            # 4. Verify executables
            # -----------------------------
            solver_0d_name = "svzerodsolver"
            solver_1d_name = "OneDSolver"
            solver_3d_name = "svmultiphysics"

            solver_0d_path = os.path.join(build_dir_0d, solver_0d_name)
            solver_1d_path = os.path.join(build_dir_1d, "bin", solver_1d_name)
            solver_3d_path = os.path.join(build_dir_3d, "svMultiPhysics-build", "bin", solver_3d_name)
            # If the build failed or never produced the exe, raise an error
            missing = []
            if not os.path.isfile(solver_0d_path):
                missing.append(solver_0d_path)
            if not os.path.isfile(solver_1d_path):
                missing.append(solver_1d_path)
            if not os.path.isfile(solver_3d_path):
                missing.append(solver_3d_path)

            if missing:
                raise RuntimeError(
                    "ERROR: Some solvers did not build correctly. Missing:\n  "
                    + "\n  ".join(missing)
                )

            # --------------------------------
            # 5. Copy executables into package
            # --------------------------------
            solvers_dir = os.path.join("svv", "solvers")
            os.makedirs(solvers_dir, exist_ok=True)
            try:
                shutil.copy2(solver_0d_path, os.path.join(solvers_dir, solver_0d_name))
                shutil.copy2(solver_1d_path, os.path.join(solvers_dir, solver_1d_name))
                shutil.copy2(solver_3d_path, os.path.join(solvers_dir, solver_3d_name))
            except Exception as e:
                raise RuntimeError("Failed copying solver executables to package directory.") from e

            # ----------------------------------
            # 6. CLEANUP if everything succeeded
            # ----------------------------------
            # We'll remove archives + source folders now that we have the built executables.
            try:
                # Remove the archives
                if os.path.isfile(tarball_path_0d):
                    os.remove(tarball_path_0d)
                if os.path.isfile(tarball_path_1d):
                    os.remove(tarball_path_1d)
                if os.path.isfile(tarball_path_3d):
                    os.remove(tarball_path_3d)

                # Remove the source directories
                if os.path.isdir(source_path_0d):
                    shutil.rmtree(source_path_0d, ignore_errors=True)
                if os.path.isdir(source_path_1d):
                    shutil.rmtree(source_path_1d, ignore_errors=True)
                if os.path.isdir(source_path_3d):
                    shutil.rmtree(source_path_3d, ignore_errors=True)

                # Remove the temporary build directories
                if os.path.isdir(build_dir):
                    shutil.rmtree(build_dir, ignore_errors=True)

                print("Cleanup complete: Removed source folders, downloaded archives, and temporary build directories.")
            except Exception as e:
                # Usually you might just print a warning. We'll raise to be explicit.
                raise RuntimeError("Warning: Cleanup of archives/folders failed.") from e
            """
            super().run()
        else:
            # -----------------------------------
            # 7. Let the normal build_ext proceed
            # -----------------------------------
            super().run()

    def finalize_options(self):
        super().finalize_options()
        self.inplace = True

def get_extra_compile_args():
    extra_args = []
    if sys.platform == "win32":
        # For MSVC or clang-cl on Windows you have to choose one of the /arch options.
        cpuinfo = None
        if cpuinfo is not None:
            info = cpuinfo.get_cpu_info()
            flags = info.get("flags", [])
            # Check for AVX512 first.
            if "avx512f" in flags:
                # This flag is supported on clang-cl and may be available in your MSVC environment.
                extra_args.extend(["/O2", "/arch:AVX512"])
            elif "avx2" in flags:
                extra_args.extend(["/O2", "/arch:AVX2"])
            elif "avx" in flags:
                extra_args.extend(["/O2", "/arch:AVX"])
            else:
                # SSE2 is the minimum on most modern CPUs.
                extra_args.extend(["/O2", "/arch:SSE2"])
            update_compile_args = ['/openmp']
            update_link_args = []
        else:
            # If cpuinfo isn't available, fall back to a safe default.
            extra_args.extend(["/O2", "/arch:SSE2"])
            update_compile_args = ['/openmp']
            update_link_args = []
    else:
        # For GCC/Clang on Linux/macOS, -march=native will enable all available features.
        extra_args.extend(["-O3", "-march=native", "-ftree-vectorizer-verbose=2"])
        update_compile_args = ['-fopenmp']
        update_link_args = ['-fopenmp']
    return extra_args, update_compile_args, update_link_args

#extra_compile_args, update_compile_args, update_link_args = get_extra_compile_args()
# extra_compile_args=update_compile_args, extra_link_args=update_link_args),
extensions = [
    Extension('svv.domain.routines.c_allocate', ['svv/domain/routines/c_allocate.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.domain.routines.c_sample', ['svv/domain/routines/c_sample.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.utils.spatial.c_distance', ['svv/utils/spatial/c_distance.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.tree.utils.c_angle', ['svv/tree/utils/c_angle.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.tree.utils.c_basis', ['svv/tree/utils/c_basis.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.tree.utils.c_close', ['svv/tree/utils/c_close.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.tree.utils.c_local_optimize', ['svv/tree/utils/c_local_optimize.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.tree.utils.c_obb', ['svv/tree/utils/c_obb.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.tree.utils.c_update', ['svv/tree/utils/c_update.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.tree.utils.c_extend', ['svv/tree/utils/c_extend.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.simulation.utils.close_segments', ['svv/simulation/utils/close_segments.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
    Extension('svv.simulation.utils.extract', ['svv/simulation/utils/extract.pyx'],
              include_dirs=[numpy.get_include()], language='c++'),
]

__version__ = '0.0.32'

with open("README.md", "r", encoding="utf-8") as file:
    DESCRIPTION = file.read()

CLASSIFIERS = ['Intended Audience :: Science/Research',
               'License :: OSI Approved :: MIT License',
               'Programming Language :: Python :: 3.9',
               'Programming Language :: Python :: 3.10',
               'Programming Language :: Python :: 3.11',
               'Programming Language :: Python :: 3.12',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX :: Linux',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS']

PACKAGES = find_packages(include=['svv', 'svv.*'])

REQUIREMENTS = ["numpy<=1.24.0",
                "scipy>=1.10.1",
                "matplotlib>=3.7.5",
                "Cython>=3.0.7",
                "usearch>=2.0.0",
                "scikit-image",
                "tetgen",
                "trimesh[all]",
                "hnswlib",
                "pyvista==0.44.2",
                "scikit-learn",
                "tqdm",
                "pymeshfix==0.17.0",
                "numexpr"]

KEYWORDS = ["modeling",
            "simulation",
            "tissue-engineering",
            "3d-printing",
            "fluid-dynamics"]

setup_info = dict(
    name='svv',
    version=__version__,
    author='Zachary Sexton',
    author_email='zsexton@stanford.edu',
    license='MIT',
    python_requires='>=3.9',
    classifiers=CLASSIFIERS,
    packages=PACKAGES,
    keywords=KEYWORDS,
    description="svVascularize (svv): A synthetic vascular generation, modeling, and simulation package",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    ext_modules=cythonize(extensions),
    package_data={'svv.bin': ['*']},
    include_dirs=[numpy.get_include()],
    include_package_data=True,
    zip_safe=False,
    install_requires=REQUIREMENTS,
    cmdclass={
        'build_ext': DownloadAndBuildExt,
        'bdist_wheel': BDistWheelCmd,
    },
)

setup(**setup_info)
